/* 
 * Copyright 2010 Aalto University, ComNet
 * Released under GPLv3. See LICENSE.txt for details. 
 */
package input;

import routing.NDNRouter;
import core.DTNHost;
import core.Message;
import core.World;

/**
 * External event for creating a message.
 */
public class MessageCreateEvent extends MessageEvent {
	private int size;
	private int responseSize;
	public String data;
	
	/**
	 * Creates a message creation event with a optional response request
	 * @param from The creator of the message
	 * @param to Where the message is destined to
	 * @param id ID of the message
	 * @param size Size of the message
	 * @param responseSize Size of the requested response message or 0 if
	 * no response is requested
	 * @param time Time, when the message is created
	 */
	public MessageCreateEvent(int from, int to, String id, int size,
			int responseSize, double time, String data) {
		super(from,to, id, time, data);
		
		
		this.size = size;
		this.responseSize = responseSize;
		this.data = data;
	}

	/**
	 * Creates the message this event represents. 
	 */

/*	
	@Override
	public void processEvent(World world) {
		DTNHost to = world.getNodeByAddress(this.toAddr);
		DTNHost from = world.getNodeByAddress(this.fromAddr);			
		
		Message m = new Message(from, null, this.id, this.size, this.data);
		m.setResponseSize(this.responseSize);
		m.setTtl(500);
		from.createNewMessage(m);
	} 
	 
*/
	
	/**
	 * Creates the message this event represents. 
	 */
	@Override
	public void processEvent(World world) {
		DTNHost to = world.getNodeByAddress(this.toAddr);
		DTNHost from = world.getNodeByAddress(this.fromAddr);
		NDNRouter fromRouter = (NDNRouter)from.getRouter();
		
		Message m = new Message(from, null, this.id, this.size, this.data);
		m.setResponseSize(this.responseSize);
		from.createNewMessage(m);
	
		m.setTtl(600);
		fromRouter.insertPI(m, true, null); // insert Pending Interest in PIT 
			// criar mtodo que recebe a mensagem 
	}

	
/*
	@Override
	public void processEvent(World world) {
		DTNHost to = world.getNodeByAddress(this.toAddr);
		DTNHost from = world.getNodeByAddress(this.fromAddr);
		sctrRouter fromRouter = (sctrRouter)from.getRouter();
		
		Message m = new Message(from, null, this.id, this.size, this.data);
		m.setResponseSize(this.responseSize);
		from.createNewMessage(m);
	
		fromRouter.insertPI(m, true, null);
		
		// criar mtodo que recebe a mensagem 
	}
	*/

	
	@Override
	public String toString() {
		return super.toString() + " [" + fromAddr + "->" + toAddr + "] " +
		"size:" + size + " CREATE";
	}
}
