package input;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;
import java.util.TreeMap;


public class MessagesGenerator {

	
	static String msgPrefix = "M";
	int max = 8;
	int min = 2;

	static int end = 86400; // 24h
	static double start = 0; 

	
	/** Time range for message creation (min, max) */
	protected double[] msgTime;
	/** Random number generator for this Class */
	protected  Random rng;
	
	/** Host Range*/
	protected static int[] hostRange = {0,97};
	protected static Map<Integer, Integer> lh ;
	static ArrayList< Integer >  hostsproducer;
	static ArrayList< Integer >  hostconsumer;
	
	/** List of all interests */
	private static ArrayList<String>listOfInterests;
	
	
	/**
	 * This function models a user using HTTP traffic. 
	 * Is generated a set of sessions for each consumer with a set of requests.
	 *	*/
	public static void main(String[] args) throws IOException {
		 double step = 1;
		 String data;
		 int numberConsumers = 90;
		 initListOfInterests();
		 Random gerador = new Random();
		 lh = new  HashMap<Integer, Integer>();
		int counterTotalRequests =0;
		int numberRequests =0;
		int hostto = 0;
		int msgSize;
		DecimalFormat decimal = new DecimalFormat("###.000");
		TreeMap<Double,String> lista = new TreeMap<Double, String>();
		BufferedWriter bw =  initFile("messagesEvents_98nodes_86400secondsNEW.txt");

		hostsproducer = new ArrayList< Integer >();
		hostconsumer = new ArrayList< Integer >();
		 
		/** List of all producers */
		int[] hosts=  {94,95,96,97, 22, 63, 41, 3};
		 
		for(int h : hosts) hostsproducer.add(h);
		
		for (int j=0; j<numberConsumers; j++) //generate a set of hostsConsumers
		{
			int host; 
			do {
				 host = gerador.nextInt(hostRange[1]-hostRange[0]) + hostRange[0];
			 } while(hostconsumer.contains(host) || hostsproducer.contains(host));
			 
			hostconsumer.add(host);
		}

		initListHostSessions();
		 
		
		/** For each consumer host:
		 * 		->Generate a number of sessions 
		 * 		For each session:
		 * 			->Generate a number of requests 
		 * 			->Generate a interval session time
		 * 			For each request:
		 * 				->Generate a interval request time 
		 * 				->Generate a time
		 * 				->Generate a size of message
		 * 				->Generate a random Interest
		 *			*/
		
		for(int i : hostconsumer) // for each consumer host
		{
			if(hostsproducer.contains(i)) continue;
			int nrsessions =0;
			double intersessions =0;
			
			if(lh.containsKey(i)){
				
				nrsessions = lh.get(i);
			}
			else continue;
			
			double lastTime = 0;
			
			for (int j =0; j<nrsessions; j++) // for each session
			{
				hostto = drawHostId(i);
				intersessions = drawTimeBetweenSession(end, lh.get(i)); // draw interval session time
				numberRequests = drawNumberRequestesPerSession(); // draw number of requests per session
				double time = 0;
				time = intersessions + lastTime;
				
				if(time>=end) continue ;
					
				for (int k=0; k<numberRequests; k++) // for each request
				{
					step = drawInterRequestsTime(0.001,0.73); // interval request time
					msgSize = (int)drawRequestSize(); // size of message of interest
					time = time + step;
					String nameInt = getRandInterest(); // get the name of interest
					data = decimal.format(time) +"	" + "C" +"	"+msgPrefix + counterTotalRequests +"	" + i + "	"+  hostto  +"	" + msgSize +"	" + nameInt;
					System.out.println(data); 
					
					lista.put(time, data);
					counterTotalRequests = counterTotalRequests+1;
				}
				
				lastTime = time;
	
			}
		}
		
		Map<Double, String> treeMap = new TreeMap<Double, String>(lista);
		Collection<String> c = treeMap.values();
		
		//Write all interests in the file
		for (String d : c)
		{
			bw.write(d);
			bw.write("\n");
		}
		bw.flush();
		bw.close();
	
	}
	
	
	/** Starts the file for writing.
	 * @param path file location 
	 **/
	public static BufferedWriter initFile(String path)
	{
		File file = new File(path);
		// if file doesnt exists, then create it
		if (!file.exists()) {
			try {
				file.createNewFile();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		 
		FileWriter fw = null;
		try {
			fw = new FileWriter(file.getAbsoluteFile());
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		BufferedWriter bw = new BufferedWriter(fw);
		return bw;
	}

	
	/**Calculates a time interval between requests.
	 * @param min is the lower limit of the range
	 * @param max is the maximum limit of the range
	 **/
	public static double drawInterRequestsTime(double min, double max)
	{
		
		double r = ((1 - Math.random()) * (max - min) + min);
		
	    if (r <= 0.73) {
			   
			   if (r < 0.15) 
				    
					return Math.pow(130, (-6));

				else if (r < 0.27)  
					
					return Math.pow(50, (-3));
					
				else if (r < 0.37) 
					
					return 1;
					
				else if (r < 0.45)  
					
					return 10;
				
				else if(r<0.52) 
					
					return 5;
			   
				else if (r<0.58)
					
					return 60;
			   
				else if (r<0.63)
						
					return Math.pow(10, (-3));
			   
				else if (r<0.67)
							
					return Math.pow(2, (-3));
							
				else if (r<0.7)
							
					return Math.pow(800, (-6));

		   }
		 
		return 0;
		
	}
	
	
	
	/**Draw number of sessions per node.
	 * @param i is the time of simulation
	 * @return the number of sessions 
	 **/
	public static int drawNumberSessionsPerNode(int i)
	{
		
		double r =0;
	
		if (i<=26000) {
			 r = ((1 - Math.random()) * (4 - 1) + 1);
			 return (int)r;
		}
		
		if (i<=43200) {
			 r = ((1 - Math.random()) * (6 - 3) + 3);
			 return (int)r;
		}
		
		if (i<=86400) {
			 r = ((1 - Math.random()) * (8 - 5) + 8);
			 return (int)r;
		}
		
		
		if (i<=172800) {
			 r = ((1 - Math.random()) * (12 - 5) + 5);
			 return (int)r;
		}
		
		if (i<=259200) {
			 r = ((1 - Math.random()) * (18 - 6) + 6);
			 return (int)r;
		}
		
		if (i<=345600) {
			 r = ((1 - Math.random()) * (26 - 12) + 12);
			 return (int)r;
		}
		
		return (int)r;
		
	}
	
	

	/**Draw number of requests per session.
	 * @return the number of requests 
	 **/
	public static  int  drawNumberRequestesPerSession()
	{
		double r = Math.random();
		
		
		   /*if (r < 0.5) {
			  System.out.println("2");
			  return 2;
		   }
		   else if (r < 0.69)
		   {
			   System.out.println("5");
			   return 5;
		   }
		   else if (r<0.79)
		   {
			   System.out.println("9");
			   return 9;
			   
		   }
		   else if ( r < 0.88)
		   {
			   System.out.println("40");
			   return 20; // alterei
		   }
		   else if (r<0.94)
		   {
			   System.out.println("80");
			   return 40; //alterei
			   
		   }
		   else if (r<0.97)
		   {
			   System.out.println("100");
			   return 83; // alterei
		   }
		   else if (r<0.99)
		   {
			   System.out.println("500");
			   return 200; // alterei
		   }
		   else return 500;*/
		
		   if (r < 0.5) {
			  return 2;
		   }
		   else if (r < 0.69)
		   {
			   return 5;
		   }
		   else if (r<0.79)
		   {
			   return 9;
		   }
		   else if ( r < 0.87)
		   {
			   return 40;
		   }
		   else if (r<0.93)
		   {
			   return 80;
		   }
		   else if (r<0.96)
		   {
			   return 100;
		   }
		   else if (r<0.98)
		   {
			   return 300;
		   }
		   else return 700;

	}
	
	/**Draw the size of request
	 * @return size of message of request
	 **/
	protected static double drawRequestSize()
	{
		double rand = Math.random();
		
		// 50 percent of requests are between 700 and 800 bytes
		// request header size is between 200 - 2kb 
		int size;
		
		if (rand < 0.5)
		{
			size = (int) ((1 - Math.random()) * (800 - 700) + 700);
		}
		else 
			{
			size = (int) ((1 - Math.random()) * (2000 - 200) + 200);
			}
		return size;
	}
	
	
	
	/**Generate randomly one host ID 
	 * @return host ID
	 **/
	protected static int drawHostId(int from)
	{
		double host;
		   
		do {
			
			host = (int) ((1 - Math.random()) * (hostRange[1] - hostRange[0]) + hostRange[0]);
			
		}while(host == from && host > hostRange[1]);
			   
		return (int) host;

	}
	
	
	protected static void initListHostSessions()
	{
		for (int i : hostconsumer)
		{
			if(hostsproducer.contains(i)) continue;
			lh.put(i, drawNumberSessionsPerNode(end));
		}
	}
	
	
	/**Draw time between Sessions
	 * @param time of simulation
	 * @param nrsession is the number of sessions for one host
	 * @return the time between sessions
	 **/
	protected static int drawTimeBetweenSession(double time, int nrsessions)
	{
		double max = time/nrsessions;
		int size = (int) ((1 - Math.random()) * (max - 0) + 0);
		return size;
	}
	

	/**Draw time between Sessions
	 * @param time of simulation
	 * @param nrsession is the number of sessions for one host
	 * @return the time between sessions
	 **/
	public static void initListOfInterests()
	{
		listOfInterests = new ArrayList<String>();
		String list_of_interests = "lista_de_interesses.txt";
		String interest;
		try{
			File arquivo = new File(list_of_interests);
			FileReader fr = new FileReader(arquivo);
			BufferedReader br = new BufferedReader( fr );
			while(br.ready()){
				interest = br.readLine();
				listOfInterests.add(interest);	
			}
			br.close();
			fr.close();
		}
		catch(Exception e){
			System.out.println("Error" + e + " in reading " + list_of_interests);
		}
	}

	/**obtaining a random interest
	 * @return name of interest
	 **/
	public static String getRandInterest(){
		Random r = new Random();
		String interest = listOfInterests.get(r.nextInt(listOfInterests.size()));
		return interest;
	}


}



