/* 
 * Copyright 2010 Aalto University, ComNet
 * Released under GPLv3. See LICENSE.txt for details. 
 */
package report;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.Collection;
import java.util.List;

import core.DTNHost;
import core.Message;
import core.MessageListener;
import core.SimClock;
import core.SimScenario;

/**
 * Reports information about all created messages. Messages created during
 * the warm up period are ignored.
 * For output syntax, see {@link #HEADER}.
 */
public class CreatedMessagesReport extends Report implements MessageListener {
	public static String HEADER = "# time  ID  size  fromHost  toHost  TTL  " + 
		"isResponse";

	protected static File file;
	static FileOutputStream outFile;
	static PrintWriter out;
	
	// TODO: This must be reviewed! Filenames can not be hardcoded HERE!
	// Frowarding Strategy
	String filename = "NDNReport_ForwardingStrategy_CreatedMessages.dat";

	
	// String filename = "C:\\Users\\Paulo\\Documents\\MATLAB\\ForwardingStrategy\\CreatedMessagesReport.dat";
	
	//PeopleShareWithLAC
	//String filename = "C:\\Users\\Paulo\\Documents\\MATLAB\\PeopleShare\\CreatedMessagesReport.dat";
	
	//String filename = "C:\\Users\\Paulo\\Documents\\MATLAB\\NDDTNProactive\\CreatedMessagesReport.dat";
	
	//String filename = "C:\\Users\\Duarte\\workspace\\ONE\\src\\ReportsFiles\\NDDTN_GROWING\\CreatedMessagesReport.dat";

	//String filename = "C:\\Users\\Duarte\\workspace\\ONE\\src\\ReportsFiles\\NDDTN_ProbabilityInterests\\CreatedMessagesReport.dat";

	DecimalFormat dc = new DecimalFormat( "0.###" );
	
	public static int interests_originated = 0;
	public static int interests_deleted = 0;
	public static int data_originated = 0;
	public static int data_deleted = 0;
	public static double size_interests_originated = 0.0;
	public static double size_data_originated = 0.0;
	public static double size_interests_deleted = 0.0;
	public static double size_data_deleted = 0.0;
	
	public static double number_interests_aborted = 0.0;
	public static double number_data_aborted = 0.0;
	
	public static CreatedMessagesReport instance = null;
	public double nextime = 900; 
	
	public static double number_interests = 0.0;
	public static double number_data = 0.0;
	
	public static double sizeInterests;
	public static double sizeData;
	
	/**
	 * Constructor.
	 */
	public CreatedMessagesReport() {
		init();
	}
	
	
	public static CreatedMessagesReport getInstance() {
		
		 if(instance == null) {
	         instance = new CreatedMessagesReport();
	      }
		 return instance;
	}	
	
	@Override
	public void init() {
		super.init();
		
		file = new File(filename);
		 try {
			outFile = new FileOutputStream(file);
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		  out = new PrintWriter(outFile); 
		//write(HEADER);
	}
	
	public void wrieToFile()
	{
		
		
		List<DTNHost> nodes = SimScenario.getInstance().getHosts();
		
		double sizedata = 0;
		double sizeInterest = 0;
		
		for(DTNHost h : nodes)
		{
			
			 Collection<Message> messages= h.getMessageCollection();
			 
			 for (Message m : messages)
			 {
					String type; 
					String[] datas;
					datas = m.getData().split(" ");
					type = datas[0];
					String name = datas[1];
					int ttl = m.getTtl();
					
					if (type.equalsIgnoreCase("pi"))
					{
						number_interests ++;
						sizeInterest = sizeInterest + m.getSize();
						// incrementar numero de interesses total, rplicas e tudo
					}
					
					if (type.equalsIgnoreCase("pc"))
					{
						number_data ++;
						sizedata = sizedata + m.getSize();
						// incrementar numero de dados total, rplicas e tudo 
					}
			 }
			 
		}
		
		// Time     InteressesOriginados     TotalInteresses    Dadosoriginados    TotalDados
		 String value = dc.format(SimClock.getIntTime()) + "	" + dc.format(interests_originated) + "	"+
			 		dc.format(number_interests) +  "	" + dc.format(data_originated) +  
			 		"	" + number_data + "	" + sizeInterest + "	" + sizedata;
	 
		 out.println(value);
		 
		 number_interests = 0;
		 number_data = 0;
		 
		 
	 
	}
	
	
	
	
	



	public void newMessage(Message m) {
		if (isWarmup()) {
			return;
		}
		
	
		String type; 
		String[] datas;
		datas = m.getData().split(" ");
		type = datas[0];
		String name = datas[1];
		int ttl = m.getTtl();
		
		
		/*if (type.equalsIgnoreCase("pi") || type.equalsIgnoreCase("pc"))
		{
			String print = format(getSimTime()) + "	" + m.getId() +"	"  + type + "	" +
						  name + "	" + m.getSize() + "	" + (ttl != Integer.MAX_VALUE ? ttl : "n/a");
			
			out.println(print);
		}
		*/
		
		if (type.equalsIgnoreCase("pi"))
		{
			interests_originated = interests_originated +1;
			size_interests_deleted = size_interests_deleted + m.getSize();
		}
		
		if (type.equalsIgnoreCase("pc"))
		{
			data_originated = data_originated+1;
			size_data_originated = size_data_originated + m.getSize();
		}
		
		
		// no final posso fazer o tamanho medio de mensagens geradas mean = size_data_originated/data_originated
		
		
		
		/*	int ttl = m.getTtl();
			write(format(getSimTime()) + " " + m.getId() + " " + 
					m.getSize() + " " + m.getFrom() + " " + m.getTo() + " " +
					(ttl != Integer.MAX_VALUE ? ttl : "n/a") +  
					(m.isResponse() ? " Y " : " N "));*/
		
		if (SimClock.getTime() >= nextime)
		{
			nextime = nextime + 900;
			// chamar funao para escrever
			wrieToFile();
			
		}
		
		
	}
	
	// nothing to implement for the rest
	public void messageTransferred(Message m, DTNHost f, DTNHost t,boolean b) {
		if (SimClock.getTime() >= nextime)
		{
			nextime = nextime + 900;
			// chamar funao para escrever
			wrieToFile();
			
		}
	}
	
	
	public void messageDeleted(Message m, DTNHost where, boolean dropped) {
		
		String type; 
		String[] datas;
		datas = m.getData().split(" ");
		type = datas[0];
		
		
		if (type.equalsIgnoreCase("pi"))
		{
			interests_deleted = interests_deleted + 1;
			size_interests_originated = size_interests_originated + m.getSize();
		}
		
		if (type.equalsIgnoreCase("pc"))
		{
			data_deleted = data_deleted + 1;
			size_data_deleted = size_data_deleted + m.getSize();
		}
		
		if (SimClock.getTime() >= nextime)
		{
			nextime = nextime + 900;
			// chamar funao para escrever
			wrieToFile();
			
		}
		
	}
	
	
	
	public void messageTransferAborted(Message m, DTNHost from, DTNHost to) {
	
		String type; 
		String[] datas;
		datas = m.getData().split(" ");
		type = datas[0];
		
		
		if (type.equalsIgnoreCase("pi"))
		{
			number_interests_aborted = number_interests_aborted + 1;
			
		}
		
		if (type.equalsIgnoreCase("pc"))
		{
			number_interests_aborted = number_interests_aborted + 1;
			
		}
		
		if (SimClock.getTime() >= nextime)
		{
			nextime = nextime + 900;
			// chamar funao para escrever
			wrieToFile();
			
		}
		
	}
	
	public void messageTransferStarted(Message m, DTNHost from, DTNHost to) {
		if (SimClock.getTime() >= nextime)
		{
			nextime = nextime + 900;
			// chamar funao para escrever
			wrieToFile();
			
		}
	}

	@Override
	public void done() {
		super.done();
		System.out.println("Mean size packet data: " + size_data_originated / data_originated);
		
		out.close();
		try {
			outFile.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}
