/* 
 /* 
 * Copyright 2015 Universidade do Minho, GCOM
 * Modified by Paulo Duarte.
 * Released under GPLv3. See LICENSE.txt for details. 
 */


package routing;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.Map.Entry;

import routing.util.RoutingInfo;
import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;
import core.SimClock;
import core.Tuple;

public class NDDTN_ProbabilityInterests extends NDNRouter{
	
	public Map<String, ArrayList<DTNHost>> inPC;

	
	public NDDTN_ProbabilityInterests(Settings s) {
		
		super(s);
		
		// TODO Auto-generated constructor stub
	}
	
	public NDDTN_ProbabilityInterests(NDDTN_ProbabilityInterests r)
	{
		
		super(r);
		this.inPC = new HashMap<String, ArrayList<DTNHost>>();
		
	
	}


	@Override
	public void changedConnection(Connection con) {
		int id = 0;
		if (con.isUp()) {
			// trocar tabela de enaminhamento  
			DTNHost otherHost = con.getOtherNode(getHost());
			id = 0;
			Message m = null;
			updateDeliveryPredFor( otherHost); 
			updateTransitivePreds(otherHost);
		}
		
	}
	
	

	@Override
	public Message messageTransferred(String id, DTNHost from) {
		Message m = super.messageTransferred(id, from); 
		String data = m.getData();
		
		/** 
		 * datas[0] = type
		 * datas[1] = name
		 * datas[2] = content
		 */
		
		String[] datas = data.split(" ");
		
//		System.out.println( "" + this.host + " recebeu-> " + datas[0]+ " -> " + datas[1] + " de: " + m.getFrom());
		
		
		//Caso receba PI
		if(datas[0].equalsIgnoreCase(INTEREST_PACKAGE)){	
			calcpopularity();
			onInterest(m, datas[1], from);
		}
		
		//Caso receba PC
		if(datas[0].equalsIgnoreCase(CONTENT_PACKAGE)){
			calcpopularity();
			onData(m, datas[1], from);
		}
		
	

		return m;
	}
	
	
	public void calcpopularity()
	{
		if(this.nrdatasent>0 && this.nrinterestsreceived>0)
		{
			double datasent = (double)this.nrdatasent;
			double intreceived = (double)this.nrinterestsreceived;
			this.popularity =(datasent/intreceived);
		}
		
	}
	
	/**
	 * Updates delivery predictions for a host.
	 * <CODE>P(a,b) = P(a,b)_old + (1 - P(a,b)_old) * P_INIT</CODE>
	 * @param host The host we just met
	 */
	private void updateDeliveryPredFor(/*Message msgCI,*/ DTNHost to) {
		
		// chamar mtodo para trocar pred
		Double oldValue = 0.0;
		MessageRouter otherRouter = to.getRouter();	
		Map<String, Integer> otherContents = ((NDDTN_ProbabilityInterests)otherRouter).lc;
		
		for (String name : otherContents.keySet())
		{
			// ver o delivered messages se tem o conteudo 
			
			if (!(this.bloomLC.contains(name)) && !(checkifExistesDeliveredmessages(name)))
			{
			 oldValue = getPredFor(name);
			 if (oldValue == null) oldValue = 0.0;
			 
			 double newValue = oldValue + (1 - oldValue) * P_INIT; 
			 
			 this.FIB.put(name, newValue);
			// removeFromMessages(m.getId());
			 
				if(this.bloomPIT.contains(name)) 
				{
					for (Message m : this.getMessageCollection()) {
						 String datas [] = m.getData().split(" ");
						 
						if (to.getRouter().hasMessage(m.getId())) {
							continue; // skip messages that the other one has
						}
						 
						 if (datas[0].equalsIgnoreCase("pi"))
						 {
							 if (datas[1].equalsIgnoreCase(name))
							 {
								 //envia interesse logo 
								
								if( sendMessage(m.getId(), to) <1){ // aqui define o tempo que foi entregue a mensagem
								// if(this.li.containsKey(m.getId()))
								//	 this.li.get(m.getId()).setoutTimeInterest(SimClock.getTime());
	
								 // Aqui no quer dizer que enviou logo o interesse !!!
								 
									if(I.LI.containsKey(m.getId()))
									{
										if(!(I.getisSendOutInterest(m.getId()))){
										 I.setTimeSendOutInterest(m.getId());
										 I.setisSendOutInterest(m.getId(), true);								 
										 I.incrSendOutInterests();
									 } 
								 }
								}
									 
							 }
						 }
					}
				
				}	
				
			}
		}
		return;
	}

	/**
	 * Updates transitive (A->B->C) delivery predictions.
	 * <CODE>P(a,c) = P(a,c)_old + (1 - P(a,c)_old) * P(a,b) * P(b,c) * BETA
	 * </CODE>
	 * @param host The B host who we just met
	 */
	
	private void updateTransitivePreds(DTNHost host) {
		
		MessageRouter otherRouter = host.getRouter();
		double pa_c_old =0;
		
		//obter delivery pred para o host a que est conetado
		// ir a fib e ver se existe alguma entrada para aquele host ja criada 
		// obter o valor 
		
		
		//aqui meter sempre o valor de default
		
		double  pa_b = 0.75;// getPrev(host); // neste mtodo tenho de arranjar forma de saber qual a prev mais alta que aquele host tem
										  // desta forma sei que ele  o produtor de tais conteudos 
										  // ou ento deixar de lado para ja isto e passar a estrategia propriamente dita
										  // receo de PI, Conteudo, PIT, FIB, analisar a memria a ver se aceita mais conteudo 
										  // tratamento da PIT para ver se aceita novos interesses 
		
		Map<String, Double> otherPreds = 
			((NDDTN_ProbabilityInterests)otherRouter).getDeliveryPreds();
		
		Map<String, Integer> otherContent = 
				((NDDTN_ProbabilityInterests)otherRouter).lc;
		
		for (String e : otherPreds.keySet()) {
			
			if (this.bloomLC.contains(e) || checkifExistesDeliveredmessages(e) || otherContent.containsKey(e)) // verifica se no contm nenhum contedo do qual ele possui
			{
				continue;
			}
			
	
			// vai existir redundancia
			// o no passa a ter os dados armazenados e tambm passa a ter os dados na FIB -.-
			// se calhar quando receber os dados, elimina da FIB o valor
			
			double pb_c = 0;
			
			pb_c = otherPreds.get(e);

			pa_c_old = getPredFor(e); // P(a,c)_old
	
			double pNew = pa_c_old + ( 1 - pa_c_old) * pa_b * pb_c * beta; //p(a,c) = P(a,c)old + (1-P(a,c)old).P(a,b).P(b,c)xBeta
		
			this.FIB.put(e, pNew);
			
			if(!this.bloomFIB.contains(e)) {
				this.bloomFIB.add(e);
			}
			
			
		}
	}

	
	
	/**
	 * Returns the current prediction (P) value for a host or 0 if entry for
	 * the host doesn't exist.
	 * @param host The host to look the P for
	 * @return the current P value
	 */
	
	public double getPredFor(String name) {
		ageDeliveryPreds(); // make sure preds are updated before getting
		if (this.FIB.containsKey(name)) {
			return this.FIB.get(name);
		}else
		{
			return 0;
		}
		
	}
	


	/**
	 * Ages all entries in the delivery predictions.
	 * <CODE>P(a,b) = P(a,b)_old * (GAMMA ^ k)</CODE>, where k is number of
	 * time units that have elapsed since the last time the metric was aged.
	 * @see #SECONDS_IN_UNIT_S
	 */
	private void ageDeliveryPreds() {
		double timeDiff = (SimClock.getTime() - this.lastAgeUpdate) / 
			secondsInTimeUnit;

		Double value; 
		
		if (timeDiff == 0){
			this.lastAgeUpdate = SimClock.getTime();
			return;
		}
		
		double mult = Math.pow(GAMMA, timeDiff);
		
		if (this.FIB.isEmpty()){
				return;
		}
		
		for (String e : this.FIB.keySet()){
			value = this.FIB.get(e);
			this.FIB.put(e, mult*value);  
		}
		
		this.lastAgeUpdate = SimClock.getTime();
	}
	
	/**
	 * Returns a map of this router's delivery predictions
	 * @return a map of this router's delivery predictions
	 */
	private HashMap<String, Double> getDeliveryPreds() {
		ageDeliveryPreds(); // make sure the aging is done
		return this.FIB;
	}
	
	// onData
	// onInterest
	// onFIBDigest
	// TrySendOutInterest
	// TrySendOutData
	// TrySendOutFIBDigest
	// receivedDuplicatedInterest
	// updateDeliveryPredsFor
	// updateTransitivyPredsFor
	// SatisfyPendingInterest
	
	
	// gesto de mensagens onde elimina as mensagens de interesse que foram criadas para o mesmo contedo
	// recebeu contedo e se tiver mensagens de interesse para esse contedo so eliminadas e dadas como satisfeitas
	
	public String removeInterest(String name, boolean setSatisfyPendingInterest)
	{
		Collection<Message> msgCollection = getMessageCollection();
		
		for (Message m : msgCollection) {
			String id;
			String data = m.getData();
			String datas[] = data.split(" ");
			
			
			if(datas[1].equalsIgnoreCase(name)){
				if(datas[0].equalsIgnoreCase(INTEREST_PACKAGE)){
					id = m.getId();
					
				//	if (I.LI.containsKey(m.getId())){
						removeFromMessages(m.getId());
						
//						System.out.println("Recebeu dados -> Removeu Interesse + " +name );
						
						// aqui depois podia combinar se ele era subscritor ou no
						if(setSatisfyPendingInterest){ // se true informa que o interesse foi satisfeito
							//this.li.get(m.getId()).setsastifiedTimeInterest(SimClock.getTime());
							I.incrSastifiedInterests();
							R.nrDeliveredData(); 
						}// incrementa numero de dados entregues
					    return id;
					//}
				}
			}
			
		}
		return null;	
	}
	
	
	/**
	 * Reception of a data packet that is treated
	 * @param m is message
	 * @param name is the name of content
	 * @param from is the sender
	 */
	
	// com CS oportunista 
	
	/*public void onData(Message m, String name, DTNHost from){	
		
		ArrayList<DTNHost> allHosts = new ArrayList<DTNHost>();
		
		if (bloomPIT.contains(name))
		{
			if (this.PIT.containsKey(name))
			{
				allHosts = this.PIT.get(name).getAllHosts();
				
				// esta parte  para encaminhar os dados recebidos pelo mesmo caminho que recebeu interesse
				
				for (DTNHost host: allHosts)
				{
					if(this.getHost() == host) continue;
					else 
					{
						if(this.lei.containsKey(name))
						 {
							 if(!this.lei.get(name).contains(host))
								 this.lei.get(name).add(host);
						 }
						 else {
							 ArrayList<DTNHost> aux = new ArrayList<DTNHost>();
							 aux.add(host);
							 this.lei.put(name, aux);
						 }
						
					}
				}
				

				if(allHosts.contains(this.getHost())) // se o n for subscritor do conteudo tbm
				{	
					if (lc.containsKey(name))
					{
							if (this.lc.get(name)!=1) this.lc.put(name, 2); // adiciona a CS o conteudo
					}else  this.lc.put(name, 2); // adiciona a CS o conteudo
					
														 
					System.out.println("No:" + this.getHost() + "recebeu dados: " + name + "de: " + m.getFrom()); 
					
					System.out.println("Armazena mensagem ");
					
					//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
					
					// Significa que agora j tem na LC o contedo 
					// para os outros ns no calcularem a transitividade mas sim a presivilidade -.-
					if (this.bloomFIB.contains(name))
					{
						this.bloomFIB.remove(name);
						this.FIB.remove(name);
					}
					
					deliveredMessages.put(m.getId(), m);
					
					//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
					
					
					//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
					// 	Se no tiver outras interfaces para enviar (faces)
					if(allHosts.size() <= 1)
						removeFromMessages(m.getId());
					else m.setFrom(this.getHost());
					
					// remove da PIT o nome do interesse
					String idm[] = m.getId().split("_");
					bloomPIT.remove(name);
					this.PIT.remove(name);	
					removeInterest(name, true);
					
					
					// RELATRIOS
					// define o tempo de chegada de dados ou pode-se dizer que  quando o interesse  satisfeito
					if (I.LI.containsKey(idm[1])){
						I.settimeSastifiedInterest(idm[1]);
						I.setisSatisfiedInterest(idm[1], true);
					}
					
					// criar lista local que serve para guardar o instante de quando  enviado o primeiro e  satisfeito
					
					R.timeDelivered(m.getId());
					R.setDeliveredData(m.getId(),  true);
					R.nrofhops = m.getHopCount();
				}
				else 
				{			
					m.setFrom(this.getHost());
					System.out.println("Armazena Dados em Buffer com id: " + m.getId());
					if(!this.lc.containsKey(name)) this.lc.put(name, 3); // adiciona a CS o conteudo
					this.bloomLC.add(name);
				
					// remove interesse pois ja recebeu dados 
					bloomPIT.remove(name);
					
					if (bloomFIB.contains(name))
					{
						bloomFIB.remove(name);
						this.FIB.remove(name);
					}
						
					this.PIT.remove(name);
					removeInterest(name, false);
					//agora devia preparar-se para enviar os dados para as interfaces que pediram 
				}
			}
			else {
				this.PIT.remove(name);
				bloomPIT.remove(name);
				removeInterest(name, false);
				m.setTtl(20);
				
			}
		}
		m.setTtl(20); //  como se fosse eliminado

	}*/
	
public void onData(Message m, String name, DTNHost from){	
		
		ArrayList<DTNHost> allHosts = new ArrayList<DTNHost>();
		boolean isubscriber = false;
		
		if (bloomPIT.contains(name))
		{
			if (this.PIT.containsKey(name))
			{
				allHosts = this.PIT.get(name).getAllHosts();
				
				// esta parte  para encaminhar os dados recebidos pelo mesmo caminho que recebeu interesse
				
				for (DTNHost host: allHosts)
				{
					if(this.getHost() == host)
						{
							if (lc.containsKey(name))
							{
									if (this.lc.get(name)!=1) this.lc.put(name, 2); // adiciona a CS o conteudo
							}else  this.lc.put(name, 2); // adiciona a CS o conteudo
							
	
							deliveredMessages.put(m.getId(), m);
							
							//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
							
							
							//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
							// 	Se no tiver outras interfaces para enviar (faces)
							if(allHosts.size() <= 1)
								removeFromMessages(m.getId());
							else m.setFrom(this.getHost());
							
							// remove da PIT o nome do interesse
							String idm[] = m.getId().split("_");
							bloomPIT.remove(name);
							this.PIT.remove(name);	
							removeInterest(name, true);
							
							
							// RELATRIOS
							// define o tempo de chegada de dados ou pode-se dizer que  quando o interesse  satisfeito
							if (I.LI.containsKey(idm[1])){
								I.settimeSastifiedInterest(idm[1]);
								I.setisSatisfiedInterest(idm[1], true);
							}
							
							// criar lista local que serve para guardar o instante de quando  enviado o primeiro e  satisfeito
							
							R.timeDelivered(m.getId());
							R.setDeliveredData(m.getId(),  true);
						//	R.nrofhops = m.getHopCount();
							
							isubscriber = true;
						}
						else 
						{
							if(this.lei.containsKey(name))
							 {
								 if(!this.lei.get(name).contains(host))
									 this.lei.get(name).add(host);
							 }
							 else {
								 ArrayList<DTNHost> aux = new ArrayList<DTNHost>();
								 aux.add(host);
								 this.lei.put(name, aux);
							 }
						
					}
				}
				
				if(isubscriber!=true)
				{
					m.setFrom(this.getHost());
					//System.out.println("Armazena Dados em Buffer com id: " + m.getId());
					if(!this.lc.containsKey(name)) this.lc.put(name, 3); // adiciona a CS o conteudo
					this.bloomLC.add(name);
				
					// remove interesse pois ja recebeu dados 
					bloomPIT.remove(name);
					this.PIT.remove(name);
					

					removeInterest(name, false);
					//agora devia preparar-se para enviar os dados para as interfaces que pediram 
				}
			}

		}
		else { 
				if (this.getFreeBufferSize() > m.getSize())
				{
					//se tiver alta taxa de entrega de dados ao consumidor
					calcpopularity();
					
					if(!this.lc.containsKey(name)) this.lc.put(name, 3); // adiciona a CS o conteudo
					if(!this.bloomLC.contains(name)) this.bloomLC.add(name);
				}
		}
		

	}
	
	public void sendDataMessage(Message c, String name, NDNRouter othRouter)
	{			
		 if(this.lei.containsKey(name))
		 {
			 if(!this.lei.get(name).contains(othRouter.getHost()))
				 this.lei.get(name).add(othRouter.getHost());
		 }
		 else {
			 ArrayList<DTNHost> aux = new ArrayList<DTNHost>();
			 aux.add(othRouter.getHost());
			 this.lei.put(name, aux);
		 }
		
		Tuple<Message, Connection> s = trySendData();
				
	}
	

	/**
	 * Reception of a interst packet that is treated. this time the packet is analyzed
	 * and the conditions for acceptance of new interest are checked. 
	 * If any content that satisfies the interest is returned immediately, 
	 * otherwise the PIT is added a new entry with the conditions.
	 * @param m is message
	 * @param name is the name of content
	 * @param from is the sender
	 */
	
	public void onInterest(Message m, String name, DTNHost from)
	{	
		String content = "";
		String[] data = name.split("/");
		int size = data.length;
		content = data[size-1];
		NDNRouter othRouter = (NDNRouter)from.getRouter();
		Message c = null;
		
		if (this.bloomLC.contains(name)) // se existir conteudo enviar 
		{
		//	content = this.lc.get(name);
		//	ystem.out.println("recebeu interesse do " +  from +" " + name);
			this.nrinterestsreceived = this.nrinterestsreceived + 1;
			// ver se no existe nenhuma mensagem para este n 
			
			if(!this.hasMessage(RESPONSE_PREFIX+m.getId()) || !this.deliveredMessages.containsKey(RESPONSE_PREFIX+m.getId()))
			{
				c = new Message(this.getHost(), othRouter.getHost(), RESPONSE_PREFIX+m.getId(), drawMessageSize(), CONTENT_PACKAGE+ " " + name + " " +name);

				
				if(createNewMessage(c)){
					//send PC
					c.setTtl(500);
					sendDataMessage(c, name,  othRouter);
					//removeFromMessages(m.getId());
				}
					 
			}
			else {
				if(!this.hasMessage(RESPONSE_PREFIX+m.getId()))
				{
					if (this.deliveredMessages.containsKey(RESPONSE_PREFIX+m.getId()))
					{
						Message msg = this.deliveredMessages.get(RESPONSE_PREFIX+m.getId());
						msg.setFrom(this.host);
						msg.setTo(othRouter.getHost());
						addToMessages(msg, false);
						sendDataMessage(msg, name, othRouter);
					}
				}
				else 
				{
					Collection<Message> msgCollection = getMessageCollection();
					
					for(Message d : msgCollection )
					{
						if (d.getId().equalsIgnoreCase(RESPONSE_PREFIX+m.getId()))
						{
							 d.setFrom(this.host);
							 d.setTo(othRouter.getHost());
							 sendDataMessage(d,  name, othRouter);
							 // m.setTtl(20);
							 //removeFromMessages(m.getId());
							 break;
						}
					}	
				}
			}
				if(I.LI.containsKey(m.getId())) 
				{
					//I.settimeSastifiedInterest(m.getId());
					I.settimeDelivered(m.getId());
				}
		
		}
		//else if (bloomPIT.contains(name)){
			//m.setTtl(20);
		//	insertPI(m, false);
		//}
		else if(bloomFIB.contains(name))
		{
			//if(comparePrevs(this, othRouter, name)>=1)
			//{
//				System.out.println();
//				System.out.println("Insere Interesse no no" + this.host);
//				System.out.println();
				
				this.nrinterestsreceived = this.nrinterestsreceived + 1;
				
				insertPI(m, false, from); // adiciona j na PIT e no  subscritor
				if (!bloomPIT.contains(name)) {
					// incrementar numero de interesses aceites
					bloomPIT.add(name);
				}
				// fica com a mensagem armazenada 
			//}
/*			else { 
				this.toremove.put(m.getId(), SimClock.getTime());
				};*/
			
		}
		else{ // se bloom FIB no possuir entrada para o interesse descarta o interesse 
			//sendInterestNack(name, m.getFrom());	
			//m.setTtl(20);			
			this.toremove.put(m.getId(), SimClock.getTime());
		}
		
	}
	
	
	
	/**
	 * Try send Dato to other router
	 * @return Tuple<Message, Connection>
	 */
	
	private Tuple<Message, Connection> trySendData() {
		
		
		List<Tuple<Message, Connection>> outgoingMessages = 
				new ArrayList<Tuple<Message, Connection>>(); 
		
			Collection<Message> msgCollection = getMessageCollection();
			
			/* for all connected hosts collect all messages that have a higher
			   probability of delivery by the other host */
			for (Connection con : getConnections()) {
				DTNHost other = con.getOtherNode(getHost());
				NDNRouter othRouter = (NDNRouter)other.getRouter();
				
				
				if (othRouter.isTransferring()) {
					continue; // skip hosts that are transferring
				}
				
				
				for (Message m : msgCollection) {

					m.setFrom(getHost()); // define o from da mensagem
					
					if (othRouter.hasMessage(m.getId())) {
						continue; // skip messages that the other one has
					}
						
					String data = m.getData();
					String[] datas = data.split(" ");
					String name = datas[1];
				
					
					if(datas[0].equalsIgnoreCase(CONTENT_PACKAGE)) {
						// send to other node
						if(this.lei.containsKey(name)) // uma forma de definir as interfaces (ns) para enviar o conteudo, neste caso quem enviou o interesse 
						{
							if(this.lei.get(name).contains(con.getOtherNode(getHost())))
							{
								outgoingMessages.add(new Tuple<Message, Connection>(m,con));
							}
						}
					}		
				} 
			}		
			if (outgoingMessages.size() == 0) {
				return null;
			}
			
			Tuple<Message, Connection> tupleSend = tryMessagesForConnected(outgoingMessages);	// try to send messages
			
			// Verifica se a mensagem foi entregue ou no 
			
			
			if(tupleSend!=null){
				
				Message send = tupleSend.getKey(); 
				if (!(R.PC.containsKey(send.getId())))
					R.createNewPC(send.getId(), (send.getData().split(" "))[1], true);
					R.incrSendOutData();
					
					Connection c = tupleSend.getValue();
					DTNHost h = c.getOtherNode(getHost());
				
					String[] dat = send.getData().split(" ");
					
					this.nrdatasent = this.nrdatasent +1;
					if(this.lei.containsKey(dat[1]))
					{
						if(this.lei.get(dat[1]).size()<=1)
						{
							this.lei.remove(dat[1]);
						}
						
					}
					
				}
			
			
			return tupleSend;
	
	}

	
	
	

		/**
		 * Try send Interest to other router
		 * @return Tuple<Message, Connection>
		 */
		private Tuple<Message, Connection> trySendInterest() {
			
			ArrayList<String> idtoRemove = new ArrayList<String>() ;
					
			List<Tuple<Message, Connection>> outgoingMessages = 
				new ArrayList<Tuple<Message, Connection>>(); 
			
			List<Tuple<Message, Connection>> out = 
					new ArrayList<Tuple<Message, Connection>>(); 
		
			Collection<Message> msgCollection = getMessageCollection();
			/* for all connected hosts collect all messages that have a higher
			   probability of delivery by the other host */
			
			for (Connection con : getConnections()) {
				DTNHost other = con.getOtherNode(getHost());
				NDNRouter othRouter = (NDNRouter)other.getRouter();
				
				
				if (othRouter.isTransferring()) {
					continue; // skip hosts that are transferring
				}
				
						
				for (Message m : msgCollection) {
					m.setFrom(getHost());
					if (othRouter.hasMessage(m.getId())) {	
						continue; // skip messages that the other one has
					}

					String split[] = m.getData().split(" ");
					String interestName = split[1];
					
					
					String data = m.getData();
					String[] datas = data.split(" ");
					String name = datas[1];
					
					if (datas[0].equalsIgnoreCase(INTEREST_PACKAGE)) {
						
						if (this.bloomLC.contains(interestName)) // Aqui o interesse  satisfeito no n
						{
							if(this.bloomPIT.contains(interestName)){
								this.bloomPIT.remove(interestName);
								this.PIT.remove(interestName);
								idtoRemove.add(m.getId());						
								I.numberInterestsSastifiedinNode();	
							}
							continue;
						}	
						else if(this.bloomPIT.contains(datas[1]))
							{
								if(othRouter.bloomLC.contains(name)) // se o outro host possuir na sua lc enviar
								{
										
									//if (this.li.containsKey(m.getId())){
										// send to other node
										
//										System.out.println("o OUTRO N " + othRouter.getHost()+ " possui dados " + name);
										out.clear();
										m.setFrom(this.getHost());
										out.add(new Tuple<Message, Connection>(m,con));
										
										Tuple<Message, Connection> res = tryMessagesForConnected(out);
										
										
										if(res!=null)
											{
												//System.out.println("ENVIOU INTERESSE!!!");
												if(!I.getisSendOutInterest(res.getKey().getId()))
												{
													I.setisSendOutInterest(res.getKey().getId(), true);
													I.setTimeSendOutInterest(res.getKey().getId());
													I.incrSendOutInterests();
												}
											}
										//outgoingMessages.add(new Tuple<Message, Connection>(m,con));
									    //}
								}
								else  if (othRouter.bloomFIB.contains(name)) // se o outro n possuir na sua FIB e a prev for > que a prev deste no enviar
								{
									if (comparePrevs(this, othRouter, name)>=1)
									{
										//if(this.li.containsKey(m.getId()))
										//{
											m.setFrom(this.getHost());
											outgoingMessages.add(new Tuple<Message, Connection>(m,con));
										//}
									}
								}	
						}
					}
				}
			}
				

			
			if (idtoRemove.size() != 0)
			{
				removemessage(idtoRemove);
			}

			
			if (outgoingMessages.size() == 0) {
				return null;
			}		
			
			// ordenar as mensages por ordem decrescente de prev a entregar 
			outgoingMessages = orderByPrev(outgoingMessages);
			// tentar enviar mensagens para as correspondentes conexes
			// retorna a mensagem e conexo para o qual foi enviada
			Tuple<Message, Connection> tupleSend = tryMessagesForConnected(outgoingMessages);
			
			if(tupleSend!=null){
				// Obtem a mensagem
				Message send = tupleSend.getKey(); 
				// Define o instante em que a mensagem foi enviada
				if(!I.getisSendOutInterest(send.getId()))
				{
					I.setisSendOutInterest(send.getId(), true);
					I.setTimeSendOutInterest(send.getId());
					I.incrSendOutInterests();
				}
			}
			
			return tupleSend; 	// try to send messages
		}

		
		/**
		 * Remove set of messages
		 * @param ArrayList Messages
		 */
		private void removemessage(ArrayList<String> messages)
		{			
			for (String Id : messages)
			{
				removeFromMessages(Id);
			}
		}
		
		/**
		 * Order Set of Messages By Probability of reach the producer
		 * @param List<Tuple<Message, Connection>>  Messages
		 * @return  List<Tuple<Message, Connection>> Messages ordered by Probability
		 */
		
		private List<Tuple<Message, Connection>> orderByPrev(List<Tuple<Message, Connection>> messages){
			
			TreeMap<String,Tuple<Message, Connection>> lista = new TreeMap<String,Tuple<Message, Connection>>();
			
			String data;
			String[] datas;
			
			int i=0;
			
			for(Tuple<Message, Connection> t : messages){
				data = t.getKey().getData();
				datas = data.split(" ");
				double prev;
				
				if (this.FIB.containsKey(datas[1]))
				{
					prev = this.FIB.get(datas[1]);
				}
				else prev=1;
				
				String pred = String.valueOf(prev+"pop"+i);
				lista.put(pred, t);
				i++;
			}
			
			List<Tuple<Message, Connection>> lista2 = new ArrayList<Tuple<Message, Connection>>();
			lista2.addAll(lista.descendingMap().values());

			return lista2;
			
		
			
		}
			
	
	


	
	/**
	 * Comparator for Message-Connection-Tuples that orders the tuples by
	 * their delivery probability by the host on the other side of the 
	 * connection (GRTRMax)
	 */
	private class TupleComparator implements Comparator  <Tuple<Message, Connection>> {

		public int compare(Tuple<Message, Connection> tuple1,
				Tuple<Message, Connection> tuple2) {
			// delivery probability of tuple1's message with tuple1's connection
			double p1 = ((ProphetRouter)tuple1.getValue().
					getOtherNode(getHost()).getRouter()).getPredFor(
					tuple1.getKey().getTo());
			// -"- tuple2...
			double p2 = ((ProphetRouter)tuple2.getValue().
					getOtherNode(getHost()).getRouter()).getPredFor(
					tuple2.getKey().getTo());

			// bigger probability should come first
			if (p2-p1 == 0) {
				/* equal probabilities -> let queue mode decide */
				return compareByQueueMode(tuple1.getKey(), tuple2.getKey());
			}
			else if (p2-p1 < 0) {
				return -1;
			}
			else {
				return 1;
			}
		}
	}
	
	@Override
	public RoutingInfo getRoutingInfo() {
		ageDeliveryPreds();
		RoutingInfo top = super.getRoutingInfo();
		RoutingInfo ri = new RoutingInfo(FIB.size() + 
				" delivery prediction(s)");
		
		for (Entry<String, Double> e : FIB.entrySet()) {
			String name = e.getKey();
			Double value = e.getValue();
			
			ri.addMoreInfo(new RoutingInfo(String.format("%s : %.6f", 
					name, value)));
		}
		
		top.addMoreInfo(ri);
		return top;
	}
	
	
	public int comparePrevs(NDNRouter r1, NDNRouter r2, String name)
	{
		double p1; // este router
		double p2; // outro router
		
		double lastProb= this.PIT.get(name).getLastProb();
		
		if (r1.FIB.containsKey(name)) p1 = r1.FIB.get(name);
		else p1 = 0;
		
		if (r2.FIB.containsKey(name)) p2 = r2.FIB.get(name);
		else p2=0;
		
		// bigger probability should come first
		if (p2-p1 == 0) {
			/* equal probabilities -> let queue mode decide */
			//return compareByQueueMode(tuple1.getKey(), tuple2.getKey());
			
			if (p2-lastProb > 0)
			{
				this.PIT.get(name).setLastProb(p2);
				return 1;
			}
			else return -1;
		}
		else if (p2-p1 < 0) {
			return -1;
		}
		else {
			if (p2-lastProb > 0)
			{
				this.PIT.get(name).setLastProb(p2);
				return 1;
			}
			else return -1;
		}		
		
	}

	
	/*public int comparePrevs(NamedDataNetworkRouterv1 r1, NamedDataNetworkRouterv1 r2, String name)
	{
		double p1;
		double p2;
		
		double lastProb= this.PIT.get(name).getLastProb();
		
		if (r1.FIB.containsKey(name)) p1 = r1.FIB.get(name);
		else p1 = 0;
		
		if (r2.FIB.containsKey(name)) p2 = r2.FIB.get(name);
		else p2=0;
		
		// bigger probability should come first
		if (p2-p1 == 0) {
			 equal probabilities -> let queue mode decide 
			//return compareByQueueMode(tuple1.getKey(), tuple2.getKey());
			return 1;
		}
		else if (p2-p1 < 0) {
			return -1;
		}
		else {
			return 1;
		}		
		
	}
*/
	
	

		
	
	@Override
	public NDDTN_ProbabilityInterests replicate() {
		return new NDDTN_ProbabilityInterests(this);
	}

	
	
	
	
	
	@Override
	public void update() {
		super.update();
		if (isTransferring() || !canStartTransfer()) {
			return; // can't start a new transfer
		}
		
	//	this.removeCImessages();
		this.trySendData();	
		this.trySendInterest();
		
	
	}

}
