package routing.NDN;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import core.Message;
import core.SimClock;

/**
 * Implementation of the results in relation to data messages
 * 
 */

public class DataReport {

	private static DataReport report = null;
	public static Map<String, Data> PC;
	public boolean initialized;
    
	public static int nrSendOutData;
	public static int nrDeliveredData;
	public static List<Integer> hopCounts;
	public static List<String> IDMessages;

	
	
	
	public class Data {
		public double timeSendOut;
		public double timeDelivered;
		public String idMessage; 
		public String Name;
		public boolean isDelivered;
		public boolean isSendOut;
		
		public Data()
		{
			timeSendOut = 0;
			timeDelivered = 0;
			idMessage = "";
			isDelivered = false;
			isSendOut = false;
		}
	}
	
	  
	public DataReport()
	{ }
	
	public static DataReport getInstance() {
		if (report == null) {
			PC = new HashMap<String, Data>();
			report = new DataReport();
			nrSendOutData =0;
			nrDeliveredData = 0;
			hopCounts = new ArrayList<Integer>();
			IDMessages = new ArrayList<String>();
		}
		return report;
	}
	
	
	/**
	 *Lets you keep a history of all data messages created and sent, maintaining their status
	 * @param IDmessage is the message ID (this ID is Unique)
	 * @param Name is the name of content
	 * @param state indicates if the message was sent or not
	 */
	public void createNewPC(String IDmessage, String Name, boolean state)
	{
		if (!PC.containsKey(IDmessage))
		{
			Data d = new Data();
			d.timeSendOut = SimClock.getTime();
			d.Name = Name; 
			d.isSendOut = state;
			PC.put(IDmessage, d);
		}
	}
	
	
	/**
	 *Increases the number of messages sent data
	 * */
	public void incrSendOutData()
	{
		nrSendOutData ++;
	}
	
	/**
	 *Increase the number of messages that satisfy interests
	 * */
	public void nrDeliveredData()
	{
		nrDeliveredData ++;
	}
	
	/**
	 *Provides the number of data messages delivered 
	 *@return number of messages delivered
	 * */
	public int getnrDeliveredData()
	{
		return nrDeliveredData;
	}
	
	/**
	 *Provides the number of data messages that were sent
	 *@return number of messages sent
	 * */
	public int getnrSendOutData()
	{
		return nrSendOutData;
	}
	
	/**
	 *Gives the instant of time that the message was delivered
	 *@param Idmessage is the message identifier
	 * */
	public void timeDelivered(String Idmessage)
	{
		PC.get(Idmessage).timeDelivered = SimClock.getTime();
	}
	
	
	/**
	 *Allows get the list of data messages
	 *@return Map with the data messages
	 * */
	public Map<String, Data> getPC()
	{	
		return this.PC;
	}
	
	
	/**
	 *Changes the state of the message
	 *@param Idmessage is the identifier of message
	 *@param s is the state of message, true if delivered
	 * */
	public void setDeliveredData(String Idmessage, boolean s)
	{
		PC.get(Idmessage).isDelivered = s;
	}
	
	
	/**
	 *Increases the number of hops. The number of hops is saved in a list
	 *@param m is the data message
	 * */
	public void incrNrHops(Message m)
	{
		this.hopCounts.add(m.getHops().size() - 1);
	}
	
	
	/**
	 *Get the number of hops
	 *@param list of the number of hops corresponding to each data message 
	 * */
	public List<Integer> getNrHops()
	{
		return this.hopCounts;
	}
	
	
	/**
	 *Records the message identifier
	 *@param id is the identifier of data message
	 * */
	public void putidMessage(String id)
	{
		if(!this.IDMessages.contains(id)) this.IDMessages.add(id);
	}
	
	
	/**
	 *Checks the message identifier exists
	 *@param id is the identifier of data message
	 * */
	public boolean existsIdMessage(String id)
	{
		if(!this.IDMessages.contains(id)) 
		{
			this.IDMessages.add(id);
			return false;
		}
		else return true;
	}
	
	

	
}
	

