package routing.NDN;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import core.Message;
import core.SimClock;

/**
 * Implementation of the results in relation to interest messages. 
 * All functions are implemented to save the results. 
 * These results are then calculated pair get statistics
 */

public class InterestReport {

	private static InterestReport report = null;
	/** List of interests*/
	public static Map<String, Data> LI;
	/** List of interests satisfied*/
	public static Map<String, Data> PITSast;
	public boolean initialized;
	/** Total number of hops*/
    public int nrofhops; 
    /** List with the corresponding number of hops for each message of interest*/
	public static List<Integer> InthopCounts;
    /** List all identifiers of interest messages*/
	public static List<String> IDMessages;
	/** Total number of messages delivered*/
	public static int nrInterestsDelivered; 
	/** Total number of interests send out*/
    public static int numberInterestsSendOut;
    /** Total number of interests satisfied*/
    public  static int numberInterestsSastified;
    /** Total number of interests satisfied locally*/
    public  static int numberInterestsSastifiedinNode;
    /** Total number of interest messages created*/
    public  static int numberInterestsCreated;
	
	
	public class Data {
		public String idMessage; 
		public String Name;
		
		public double timeSendOut; 
		public double timeDelivered; 
		public double timeCreate; 
		public double timeSatisfied;
		public boolean isDelivered;
		public boolean isSastified;
		public boolean isSendOut;
		
		public Data()
		{
			timeSendOut = 0;
			timeDelivered = 0;
			idMessage = "";
			isDelivered = false;
			isSendOut = false;
		}
	}
	
	  
	public InterestReport(){}
	
	
	public static InterestReport getInstance() {
		if (report == null) {
			LI = new HashMap<String, Data>();
			PITSast = new HashMap<String, Data>();
			report = new InterestReport();
			InthopCounts = new ArrayList<Integer>();
			IDMessages = new ArrayList<String>();
			Init();
		}
		return report;
	}
	
	
	public static  void Init()
	{
		numberInterestsSendOut = 0;
		numberInterestsSastified = 0;
		numberInterestsSastifiedinNode =0;
		numberInterestsCreated = 0;
		nrInterestsDelivered = 0;
	}
	
	
	/**
	 *Increasing the number of interests sent
	 */
	public void incrSendOutInterests()
	{
		numberInterestsSendOut++;
	}
	
	/**
	 *Increasing the number of interests satisfied
	 */
	public void incrSastifiedInterests()
	{
		numberInterestsSastified++;
	}
	
	/**
	 *Increasing the number of interests satisfied in the node
	 */
	public void numberInterestsSastifiedinNode()
	{
		numberInterestsSastifiedinNode++;
	}
	
	
	/**
	 *Create new entry in the list of interests
	 */
	public void createNewRinterest(String IDmessage, String Name)
	{
		if (!LI.containsKey(IDmessage))
		{
			numberInterestsCreated ++;
			Data d = new Data();
			d.Name = Name; 
			d.isDelivered = false;
			d.isSendOut = false;
			d.isSastified = false;
			d.timeCreate = SimClock.getTime();
			LI.put(IDmessage, d);
			
		}
		
	}
	
	
	/**
	 *Create new entry in the list of satisfied interests
	 */
	public void createPITSast(double timecreated, double timeSastified)
	{	
		Data d = new Data();
		d.timeCreate = timecreated;
		d.timeDelivered= timeSastified;
		int i= PITSast.size();	
		PITSast.put("PIT"+i, d);
	}
	
	
	/**Return the list of interests 
	 * @return  Map<String, Data> list of interests
	 */
	public Map<String, Data> getLI()
	{	
		return LI;
	}
	
	/**Set the time when the interest reach the producer/intermediately node which have data to match interest.
	 * @param Idmessage is the identifier of message e.g M01 
	 */
	
	public void settimeDelivered(String Idmessage)
	{
		if(LI.containsKey(Idmessage)) LI.get(Idmessage).timeDelivered = SimClock.getTime();
	}
	
	
	/**Set the time when the interest was send out for the first time.
	 * @param Idmessage is the identifier of message e.g M01 
	 */
	public void setTimeSendOutInterest(String Idmessage)
	{
		if(LI.containsKey(Idmessage)) LI.get(Idmessage).timeSendOut = SimClock.getTime();
	}

	
	/**Set the time when the interest was satisfied, 
	 * when the data reaches the consumer's interest is satisfied
	 * @param Idmessage is the identifier of message e.g M01 
	 */
	public void settimeSastifiedInterest(String Idmessage)
	{
		if(LI.containsKey(Idmessage)) LI.get(Idmessage).timeSatisfied = SimClock.getTime();
	}
	
	
	/**Set true if the interest was sent to the network, 
	 * is defined in the first transmission of interest
	 * @param Idmessage is the identifier of message e.g M01 
	 * @param s is the state true or false */
	public void setisSendOutInterest(String Idmessage, boolean s) 
	{
		if(LI.containsKey(Idmessage)) LI.get(Idmessage).isSendOut = s;
		
	}
	
	
	/**Returns the state of interest. If already sent out returns true.
	 *  @param Idmessage is the identifier of the message
	 */
	public boolean getisSendOutInterest(String Idmessage)
	{
		if(LI.containsKey(Idmessage)) return LI.get(Idmessage).isSendOut;
		else return false;
	}
	
	
	/**Returns the state of interest. If already satisfied returns true.
	 * @param Idmessage is the identifier of the message
	 */
	public boolean getisSastifiedInterest(String Idmessage)
	{
		if(LI.containsKey(Idmessage))  return LI.get(Idmessage).isSastified;
		else return false;
	}
	
	/**Set True if interest was Satisfied ie. if the consumer receive the data
	 * @param Idmessage is the identifier of message e.g M01 
	 * @param s is the state true or false */
	
	public void setisSatisfiedInterest(String Idmessage, boolean s)
	{
		if(LI.containsKey(Idmessage)) LI.get(Idmessage).isSastified = s;
	}
	
	
	/**Set True if interest was delivered to producer
	 * @param Idmessage is the identifier of message e.g M01 
	 * @param s is the state true or false 
	 * */
	public void setDeliveredInterest(String Idmessage, boolean s)
	{
		if(LI.containsKey(Idmessage)) LI.get(Idmessage).isDelivered = s;
	}
	
	
	/**Increase the number of hops
	 * @param m is the message 
	 * */
	public void incrNrHops(Message m)
	{
		this.InthopCounts.add(m.getHops().size() - 1);
	}
	
	
	/**Gets the number of hops
	 *@return the list of hops
	 * */
	public List<Integer> getNrHops()
	{
		return this.InthopCounts;
	}
	
	/**
	 *Records the message identifier
	 *@param id is the identifier of data message
	 * */
	public void putidMessage(String id)
	{
		if(!this.IDMessages.contains(id)) this.IDMessages.add(id);
	}
	
	
	/**
	 *Checks the message identifier exists
	 *@param id is the identifier of data message
	 * */
	public boolean existsIdMessage(String id)
	{
		
		if(!this.IDMessages.contains(id)) 
		{
			this.nrInterestsDelivered ++;
			this.IDMessages.add(id);
			return false;
		}
		else return true;
	}
	
	
	public int getnrIntDeli()
	{
		return nrInterestsDelivered;
	}
	
	
	}
	

