/* 
 /* 
 * Copyright 2015 Universidade do Minho, GCOM
 * Modified by Paulo Duarte.
 * Released under GPLv3. See LICENSE.txt for details. 
 */

package routing;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import report.NDN_Report;
import routing.NDN.PeopleShareInterest;
import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;
import core.Tuple;

public class NDN_PeopleShareRouter extends NamedDataNetworkRouter{
	
	/** PeopleShare router's setting namespace ({@value})*/ 
	public static final String NDNPS_NS = "PeopleShareRouter";

	/**
	 * Constructor. Creates a new message router based on the settings in
	 * the given Settings object.
	 * @param s The settings object
	 */
	public NDN_PeopleShareRouter(Settings s) {
		super(s);
		//Settings PeopleShareSettings = new Settings(NDNPS_NS);
	}
	
	protected NDN_PeopleShareRouter(NDN_PeopleShareRouter r) {
		super(r);
	}

	@Override
	public void changedConnection(Connection con) {
		DTNHost otherHost = con.getOtherNode(getHost());
		int id = 0;
		for(String data : this.lc.values()){
			Message m = new Message(this.getHost(), otherHost, "IC_" +id, 0 , PACOTE_INFORMACAO_CONTEUDO +" " + data);
			int lenght = getMessageLenght(m, PACOTE_INFORMACAO_CONTEUDO +" " + data);
			m = new Message(this.getHost(), otherHost, "IC_" +id, lenght , PACOTE_INFORMACAO_CONTEUDO +" " + data);
			m.setTtl(10);
			if(createNewMessage(m)){
				id++;
			}	
		}
	}

	@Override
	public Message messageTransferred(String id, DTNHost from) {
		Message m = super.messageTransferred(id, from); 
		String data = m.getData();
		/** 
		 * datas[0] = type
		 * datas[1] = name
		 * datas[2] = content
		 */
		String[] datas = data.split(" ");
		
		//System.out.println("-> " + datas[0]+ " -> " + datas[1]);
		

		//Caso receba PI
		if(datas[0].equalsIgnoreCase(PACOTE_INTERESSE)){			
			receivePI(m, datas[1], from);
		}
		//Caso receba PC
		if(datas[0].equalsIgnoreCase(PACOTE_CONTEUDO)){
			receivePC(m, datas[1], datas[2]);
		}
		
		//Caso receba PIC
		if(datas[0].equalsIgnoreCase(PACOTE_INFORMACAO_CONTEUDO)){
			receivePCD(m, datas[1], from);
		}
		return m;
	}
	
	public void receivePI(Message m, String name, DTNHost from){
		PeopleShareInterest newInterest = new PeopleShareInterest();	
		String idMsg = m.getId(); 
		String content = "";
		ArrayList<DTNHost> novaLista = new ArrayList<DTNHost>();
		Message c = null;
		
		//Se no existir na LI, adicionar novo
		NDN_Report.received_interests++;
		if(!this.li.containsKey(name)){
			addInterest(idMsg, name, 1, false, null);
			updatePopularity();
		}
		else{
			newInterest = this.li.get(name);
			newInterest.setNrofRequests((newInterest.getNrOfRequests()) + 1);
			this.li.put(name, newInterest);
			updatePopularity();
		}		
		String idInt = li.get(name).getId();
		
		//Se existir na LC enviar PC
		if(this.lc.containsKey(idInt)){ 
			content = this.lc.get(idInt);
			c = new Message(this.getHost(), from, RESPONSE_PREFIX+m.getId(), drawMessageSize(), PACOTE_CONTEUDO+ " " + name + " " +content);
			if(createNewMessage(c)){
				trySendPC();
			}			
		}
		//Seno inserir na lpi
		else{
			if(!this.lpi.containsKey(idInt)){
				novaLista.add(m.getFrom());
				this.lpi.put(idInt, novaLista);
			}
			else{
				novaLista = this.lpi.get(idInt);
				if(!novaLista.contains(m.getFrom())){
					novaLista.add(m.getFrom());
				}
				this.lpi.put(idInt, novaLista);
			}
		}
	}
	
	public void receivePC(Message m, String name, String content){
		PeopleShareInterest newInterest = new PeopleShareInterest();	
		Message c = null;
		NDN_Report.received_contents++;		
		String idInt = li.get(name).getId(); 
		newInterest = li.get(name);
		if(newInterest.getIsSubscriber()){
			NDN_Report.PC_hopCounts.add(m.getHops().size() - 1);
			NDN_Report.final_destiny++;
			this.lc.put(idInt, content);
			newInterest.setIsSubscriber(false);
			li.put(name, newInterest);
		}
		removeFromMessages(m.getId());
		if(lpi.containsKey(idInt)){
			for(int i =0 ; i< lpi.get(idInt).size(); i++){
				c = new Message(this.getHost(), lpi.get(idInt).get(i), RESPONSE_PREFIX+i+m.getId(), drawMessageSize(), PACOTE_CONTEUDO+ " " + name + " " +content);
				if(!createNewMessage(c)){
					System.out.println("No criei PC");
				}
			}
			lpi.remove(idInt);
		}
	}

	public void receivePCD(Message m, String name, DTNHost from){	
		String idMsg = m.getId(); 
		NDN_Report.received_content_pointer++;
		if(!this.li.containsKey(name)){ 
			addInterest(idMsg, name, 0, false, null);
		}		
		String idInt = li.get(name).getId();
		removeFromMessages(m.getId());
		if(this.lpi.containsKey(idInt)){
			int lenght = getMessageLenght(m,PACOTE_INTERESSE+ " " + name);
			Message r = new Message(this.getHost(), from, m.getId(), lenght, PACOTE_INTERESSE+ " " + name);
			if(createNewMessage(r)) 
				sendMessage(m.getId(),from);
		}
	}
		
	public void updatePopularity(){
		int nrOfRequests=0;
		double popularity;
		Map<String, PeopleShareInterest> temp = new HashMap<String, PeopleShareInterest>();
		temp.putAll(li);
		if(!li.isEmpty()){
			for(PeopleShareInterest i: temp.values()){
				nrOfRequests += i.getNrOfRequests(); 
			}
			for(PeopleShareInterest i: temp.values()){
				if(nrOfRequests==0){
					popularity=0;
				}
				else
					popularity = Double.valueOf(i.getNrOfRequests()) / Double.valueOf(nrOfRequests);
				i.setPopularity(popularity);
				li.put(i.getName(), i);
			}
		}
	}
	/**
	 * Tries to send all Package Content to all connected hosts ordered by
	 * their popularity
	 * @return The return value of {@link #tryMessagesForConnected(List)}
	 */
	private Tuple<Message, Connection> trySendPC() {
		List<Tuple<Message, Connection>> messages = 
			new ArrayList<Tuple<Message, Connection>>(); 
	
		Collection<Message> msgCollection = getMessageCollection();
		
		/* for all connected hosts collect all messages that have a higher
		   probability of delivery by the other host */
		for (Connection con : getConnections()) {
			DTNHost other = con.getOtherNode(getHost());
			NamedDataNetworkRouter othRouter = (NamedDataNetworkRouter)other.getRouter();
			
			
			if (othRouter.isTransferring()) {
				continue; // skip hosts that are transferring
			}
			
			for (Message m : msgCollection) {
				if (othRouter.hasMessage(m.getId())) {
					continue; // skip messages that the other one has
				}
				String data = m.getData();
				String[] datas = data.split(" ");
				if(datas[0].equalsIgnoreCase(PACOTE_CONTEUDO)) {
					if(m.getTo()==other){
						// send to other node
						messages.add(new Tuple<Message, Connection>(m,con));
					}
				}		
			}
		}		
		if (messages.size() == 0) {
			return null;
		}
		messages = orderByPopularity(messages);
		return tryMessagesForConnected(messages);	// try to send messages
	}
	
	/**
	 * Tries to send all Package Interest to all connected hosts ordered by
	 * their popularity
	 * @return The return value of {@link #tryMessagesForConnected(List)}
	 */
	private Tuple<Message, Connection> trySendPI() {
		List<Tuple<Message, Connection>> messages = 
			new ArrayList<Tuple<Message, Connection>>(); 
	
		Collection<Message> msgCollection = getMessageCollection();
		/* for all connected hosts collect all messages that have a higher
		   probability of delivery by the other host */
		for (Connection con : getConnections()) {
			DTNHost other = con.getOtherNode(getHost());
			NamedDataNetworkRouter othRouter = (NamedDataNetworkRouter)other.getRouter();
			
			
			if (othRouter.isTransferring()) {
				continue; // skip hosts that are transferring
			}
			
			for (Message m : msgCollection) {
				if (othRouter.hasMessage(m.getId())) {
					continue; // skip messages that the other one has
				}
				String data = m.getData();
				String[] datas = data.split(" ");
				
				
				if (datas[0].equalsIgnoreCase(PACOTE_INTERESSE)) {
					PeopleShareInterest newInterest = li.get(datas[1]);
					if(newInterest.getIsSubscriber()){
						// send to other node
						messages.add(new Tuple<Message, Connection>(m,con));
					}	
				}
			}			
		}
		
		if (messages.size() == 0) {
			return null;
		}		
		messages = orderByPopularity(messages);
		return tryMessagesForConnected(messages);	// try to send messages
	}

		/**
	 * Tries to send all Package Content Detail to all connected hosts ordered by
	 * their popularity
	 * @return The return value of {@link #tryMessagesForConnected(List)}
	 */
	private Tuple<Message, Connection> trySendPIC() {
		List<Tuple<Message, Connection>> messages = 
			new ArrayList<Tuple<Message, Connection>>(); 
	
		Collection<Message> msgCollection = getMessageCollection();
		
		/* for all connected hosts collect all messages that have a higher
		   probability of delivery by the other host */
		for (Connection con : getConnections()) {
			DTNHost other = con.getOtherNode(getHost());
			NamedDataNetworkRouter othRouter = (NamedDataNetworkRouter)other.getRouter();
			
			
			if (othRouter.isTransferring()) {
				continue; // skip hosts that are transferring
			}
			
			for (Message m : msgCollection) {
				if (othRouter.hasMessage(m.getId())) {
					continue; // skip messages that the other one has
				}
				String data = m.getData();
				String[] datas = data.split(" ");
						
				if (datas[0].equalsIgnoreCase(PACOTE_INFORMACAO_CONTEUDO)) {
					// send to other node
					messages.add(new Tuple<Message, Connection>(m,con));	
				}
			}			
		}
		
		if (messages.size() == 0) {
			return null;
		}
		messages = orderByPopularity(messages);
		return tryMessagesForConnected(messages);	// try to send messages
	}
	
	private List<Tuple<Message, Connection>> orderByPopularity(List<Tuple<Message, Connection>> messages){
		TreeMap<String,Tuple<Message, Connection>> lista = new TreeMap<String,Tuple<Message, Connection>>();
		String data;
		String[] datas;
		PeopleShareInterest newInterest = new PeopleShareInterest();
		int i=0;
		for(Tuple<Message, Connection> t : messages){
			data = t.getKey().getData();
			datas = data.split(" ");
			newInterest = li.get(datas[1]);	
			String popularity = String.valueOf(newInterest.getPopularity())+"pop"+i;
			lista.put(popularity, t);
			i++;
		}
		List<Tuple<Message, Connection>> lista2 = new ArrayList<Tuple<Message, Connection>>();
		lista2.addAll(lista.descendingMap().values());

		return lista2;
	}
	
	@Override
	public NDN_PeopleShareRouter replicate() {
		return new NDN_PeopleShareRouter(this);
	}
	
	@Override
	public void update() {
		super.update();
		if (isTransferring() || !canStartTransfer()) {
			return; // can't start a new transfer
		}
		this.trySendPC();
		this.trySendPI();
		this.trySendPIC();
	}
}
