/* 
 /* 
 * Copyright 2015 Universidade do Minho, GCOM
 * Modified by Paulo Duarte.
 * Released under GPLv3. See LICENSE.txt for details. 
 */

package routing;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import report.NDN_Report;
import routing.NDN.PeopleShareInterest;
import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;
import core.SimClock;
import core.Tuple;

public class NDN_PeopleShareRouter_withLAC extends NamedDataNetworkRouter{
	
	/** PeopleShare router's setting namespace ({@value})*/ 
	public static final String NDNPS_NS = "PeopleShareRouter";

	/**
	 * Constructor. Creates a new message router based on the settings in
	 * the given Settings object.
	 * @param s The settings object
	 */
	public NDN_PeopleShareRouter_withLAC(Settings s) {
		super(s);
		//Settings PeopleShareSettings = new Settings(NDNPS_NS);
	}
	
	protected NDN_PeopleShareRouter_withLAC(NDN_PeopleShareRouter_withLAC r) {
		super(r);
	}

//	@Override
//	public void changedConnection(Connection con) {
//		DTNHost otherHost = con.getOtherNode(getHost());
//		int id = 0;
//		for(String data : this.lc.values()){
//			Message m = new Message(this.getHost(), otherHost, "IC_" +id, 0 , PACOTE_INFORMACAO_CONTEUDO +" " + data);
//			int lenght = getMessageLenght(m, PACOTE_INFORMACAO_CONTEUDO +" " + data);
//			m = new Message(this.getHost(), otherHost, "IC_" +id, lenght , PACOTE_INFORMACAO_CONTEUDO +" " + data);
//			m.setTtl(10);
//			if(createNewMessage(m)){
//				id++;
//			}	
//		}
//	}

	@Override
	public Message messageTransferred(String id, DTNHost from) {
		Message m = super.messageTransferred(id, from); 
		String data = m.getData();
		/** 
		 * datas[0] = type
		 * datas[1] = name
		 * datas[2] = content
		 */
		String[] datas = data.split(" ");
		
		//System.out.println("-> " + datas[0]+ " -> " + datas[1]);
		

		//Caso receba PI
		if(datas[0].equalsIgnoreCase(PACOTE_INTERESSE)){			
			receivePI(m, datas[1], from);
		}
		//Caso receba PC
		if(datas[0].equalsIgnoreCase(PACOTE_CONTEUDO)){
			receivePC(m, datas[1], datas[2]);
		}
		
	/*	//Caso receba PIC
		if(datas[0].equalsIgnoreCase(PACOTE_INFORMACAO_CONTEUDO)){
			receivePCD(m, datas[1], from);
		}*/
		return m;
	}
	
	public void checkSatisfiedInNode()
	{
		Collection<Message> msgCollection = getMessageCollection();

		ArrayList<String> idtoRemove = new ArrayList<String>() ;
		
		for (Message m : msgCollection) {
			
			String data = m.getData();
			String[] datas = data.split(" ");
			
			if (datas[0].equalsIgnoreCase(PACOTE_INTERESSE)) {
				
				if(this.lc.containsKey(datas[1]))
				{
					if(m.getFrom() == this.getHost())
					{
						if(!idtoRemove.contains(m.getId()))
						{
							I.numberInterestsSastifiedinNode();	
						}
						else{
							System.out.println("burro");
						}
					}
					
					idtoRemove.add(m.getId());					
				}
			
			}
		}
		
		removemessages(idtoRemove);
		
	}
	
	public void receivePI(Message m, String name, DTNHost from){
		PeopleShareInterest newInterest = new PeopleShareInterest();	
		String idMsg = m.getId(); 
		String content = "";
		ArrayList<DTNHost> novaLista = new ArrayList<DTNHost>();
		Message c = null;
		
		//Se no existir na LI, adicionar novo
		NDN_Report.received_interests++;
		if(!this.li.containsKey(name)){
			addInterest(idMsg, name, 1, false, null);
			//updatePopularity();
			
			if(!I.LI.containsKey(idMsg))
			{
				I.createNewRinterest(idMsg, name);
			}
		}
		else{
			newInterest = this.li.get(name);
			newInterest.setNrofRequests((newInterest.getNrOfRequests()) + 1);
			this.li.put(name, newInterest);
			//updatePopularity();
			
			if(!I.LI.containsKey(idMsg))
			{
				I.createNewRinterest(idMsg, name);
			}
		}		
		
		
		//Se existir na LC enviar PC // Cria sempre mensagem mesmo que j tenha no buffer
		// pode acontecer que ja tenha recebido um interesse e cria na mesma na presena de mensagem de conteudo criada -.-
		if(this.lc.containsKey(name)){
			content = this.lc.get(name);
			if(this.deliveredMessages.containsKey(RESPONSE_PREFIX+m.getId()))
			{
				Message msg = this.deliveredMessages.get(RESPONSE_PREFIX+m.getId());
				msg.setTo(null);
				addToMessages(msg, true);
			}
			else if(!this.hasMessage(RESPONSE_PREFIX+m.getId()))
			{
				c = new Message(this.getHost(), null, RESPONSE_PREFIX+m.getId(), drawMessageSize(), PACOTE_CONTEUDO+ " " + name + " " +content);
				if(createNewMessage(c)){
					//R.incrSendOutData();
					// aqui posso incrementar o nmero de pacotes de dados enviados 
					if (!(R.PC.containsKey(c.getId())))
						R.createNewPC(c.getId(), (c.getData().split(" "))[1], false);
						trySendSpecificPC(c, from);
					//trySendPC();
				}		
			}else trySendPC();
			
			//else //trySendSpecificPC(this.getMessage(RESPONSE_PREFIX+m.getId()), from);
			//trySendPC();
			removeInterest(name);
		}
		//Seno inserir na lpi
		else{
			if(!this.lpi.containsKey(name)){
				novaLista.add(m.getFrom());
				this.lpi.put(name, novaLista);
			}
			else{
				novaLista = this.lpi.get(name);
				if(!novaLista.contains(m.getFrom())){
					novaLista.add(m.getFrom());
				}
				this.lpi.put(name, novaLista);
			}
		}
	}
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	public void receivePC(Message m, String name, String content){
		PeopleShareInterest newInterest = new PeopleShareInterest();	
		Message c = null;
		NDN_Report.received_contents++;		
		
		if(!this.li.containsKey(name)){
			addInterest(m.getId(), name, 1, false, null);
			//updatePopularity();
			
			if(!I.LI.containsKey(m.getId()))
			{
				I.createNewRinterest(m.getId(), name);
			}
		}
		
		
		newInterest = li.get(name);
		
		if(newInterest.getIsSubscriber()){
			
			NDN_Report.PC_hopCounts.add(m.getHops().size() - 1);
			NDN_Report.final_destiny++;
			I.incrSastifiedInterests();
			
			this.lc.put(name, content);
			
			removeInterest(name);
			
			I.createPITSast(newInterest.gettimecreated(), SimClock.getTime());
			
			newInterest.setIsSubscriber(false);
			li.put(name, newInterest);
			
			String idm[] = m.getId().split("_");
			int size = idm.length;
			
			if (I.LI.containsKey(idm[size - 1])){
				I.settimeSastifiedInterest(idm[size - 1]);
				I.setisSatisfiedInterest(idm[size -1], true);
			}
			
			if (R.PC.containsKey("R_"+idm[size - 1]))
			{
				R.timeDelivered("R_"+idm[size - 1]);
				R.setDeliveredData("R_"+idm[size - 1],  true);
				//R.nrofhops = m.getHopCount();
				R.nrDeliveredData(); 	
			}
			this.deliveredMessages.put(name, m);
			
		}else 
		{
			this.lc.put(name, content);
			removeInterest(name);
		}
		
/*		if(lpi.containsKey(name)){
			for(int i =0 ; i< lpi.get(name).size(); i++){
				c = new Message(this.getHost(), lpi.get(name).get(i), RESPONSE_PREFIX+i+m.getId(), drawMessageSize(), PACOTE_CONTEUDO+ " " + name + " " +content);
				if(!createNewMessage(c)){
					System.out.println("No criei PC");
				}
			}
			lpi.remove(name);
		}*/
		
	}

	
	
	public void removeInterest(String name)
	{
		Collection<Message> msgCollection = getMessageCollection();
		
		ArrayList <String> idtoRemove = new ArrayList<String>();
		
		for (Message m : msgCollection) {
					
			String data = m.getData();
			String[] datas = data.split(" ");
						
			if(m.getFrom() == this.getHost())
			{
				if (datas[0].equalsIgnoreCase(PACOTE_INTERESSE)) 
				{	
					if (datas[1].equalsIgnoreCase(name)) 
					{
						deleteMessage(m.getId(), true);
						break;
					}
				}
			}
		}

		
		for (Message m : msgCollection) {
			
			
			String data = m.getData();
			String[] datas = data.split(" ");
			
			if (datas[0].equalsIgnoreCase(PACOTE_INTERESSE)) 
			{	
				if(this.lc.containsKey(datas[1]))
				{						
					if(m.getFrom() == this.getHost())
					{
						if(!idtoRemove.contains(m.getId()))
						{
							I.numberInterestsSastifiedinNode();	
						}
						else
						{
							System.out.println("burro");
						}
					}
					idtoRemove.add(m.getId());		
				} 
			}
		}
		
		removemessagesaux( idtoRemove);
	}
	
	
	private Tuple<Message, Connection> trySendPC() {
		List<Tuple<Message, Connection>> messages = 
			new ArrayList<Tuple<Message, Connection>>(); 
	
		Collection<Message> msgCollection = getMessageCollection();
		
		/* for all connected hosts collect all messages that have a higher
		   probability of delivery by the other host */
		for (Connection con : getConnections()) {
			DTNHost other = con.getOtherNode(getHost());
			NamedDataNetworkRouter othRouter = (NamedDataNetworkRouter)other.getRouter();
			
			
			if (othRouter.isTransferring()) {
				continue; // skip hosts that are transferring
			}
			
			for (Message m : msgCollection) {
				if (othRouter.hasMessage(m.getId())) {
					continue; // skip messages that the other one has
				}
				String data = m.getData();
				String[] datas = data.split(" ");
				if (datas[0].equalsIgnoreCase(PACOTE_CONTEUDO)) {
					// send to other node

							messages.add(new Tuple<Message, Connection>(m,con));

				//	messages.add(new Tuple<Message, Connection>(m,con));
				}		
			}
		}
		
		if (messages.size() == 0) {
			return null;
		}
		
		Tuple<Message, Connection> tupleSend = tryMessagesForConnected(messages);	// try to send messages
		
		// Obtem a mensagem
		if(tupleSend!=null)
		{
					Message m = tupleSend.getKey();
			
					String data = m.getData();
					String[] datas = data.split(" ");
							
					if(this.li.containsKey(datas[1]))
					{
						
								
						if(this.lpi.containsKey(datas[1]))
						{
							if(this.lpi.get(datas[1]).contains(tupleSend.getValue().getOtherNode(getHost())))
							{
								this.lpi.get(datas[1]).remove(tupleSend.getValue().getOtherNode(getHost()));
							}
						}
					}
							
						
					if ((R.PC.containsKey(m.getId())))
					{
						if(!R.PC.get(m.getId()).isSendOut)
						{
							R.PC.get(m.getId()).isSendOut = true;
							R.PC.get(m.getId()).timeSendOut = SimClock.getTime();
							R.incrSendOutData();
						}	
					}
							
				}
		
		return tupleSend;
	}
	
	
	
	/**
	 * Tries to send all Package Content to all connected hosts ordered by
	 * their popularity
	 * @return The return value of {@link #tryMessagesForConnected(List)}
	 */

	
	private Tuple<Message, Connection> trySendSpecificPC(Message m, DTNHost outro) {
		List<Tuple<Message, Connection>> messages = 
			new ArrayList<Tuple<Message, Connection>>(); 
		
		ArrayList<DTNHost> novaLista = new ArrayList<DTNHost>();

		NamedDataNetworkRouter othRouter = (NamedDataNetworkRouter)outro.getRouter();
			
			
		if (othRouter.isTransferring()) {
			return null; // skip hosts that are transferring
		}
			
			
		if (othRouter.hasMessage(m.getId())) {
					return null; // skip messages that the other one has
		}
		
		String data = m.getData();
		String[] datas = data.split(" ");
		if(datas[0].equalsIgnoreCase(PACOTE_CONTEUDO)) {			
						// send to other node
				if(this.sendMessage(m.getId(), outro) == 0)
				{
					if ((R.PC.containsKey(m.getId())))
					{
						if(!R.PC.get(m.getId()).isSendOut)
						{
							R.PC.get(m.getId()).isSendOut = true;
							R.PC.get(m.getId()).timeSendOut = SimClock.getTime();
							R.incrSendOutData();
						}	
					}
				}	
				else
				{
					if(!this.lpi.containsKey(m.getData().split(" ")[1])){
						novaLista.add(m.getFrom());
						this.lpi.put(m.getData().split(" ")[1], novaLista);
					}
					else{
						novaLista = this.lpi.get(m.getData().split(" ")[1]);
						if(!novaLista.contains(m.getFrom())){
							novaLista.add(m.getFrom());
						}
						this.lpi.put(m.getData().split(" ")[1], novaLista);
					}
				}
				
				}
		return null;	
		}		
	
	
	
	
	
	public void removemessages(ArrayList<String> messages)
	{
		for (String Id: messages)
		{
			removeFromMessages(Id);
		}
		
		
	}
	
	
	private Tuple<Message, Connection> trySendPI() {
		List<Tuple<Message, Connection>> messages = 
			new ArrayList<Tuple<Message, Connection>>(); 
	
		Collection<Message> msgCollection = getMessageCollection();
		/* for all connected hosts collect all messages that have a higher
		   probability of delivery by the other host */
		for (Connection con : getConnections()) {
			DTNHost other = con.getOtherNode(getHost());
			NamedDataNetworkRouter othRouter = (NamedDataNetworkRouter)other.getRouter();
			
			
			if (othRouter.isTransferring()) {
				continue; // skip hosts that are transferring
			}
			
			for (Message m : msgCollection) {
				if (othRouter.hasMessage(m.getId())) {
					continue; // skip messages that the other one has
				}
				String data = m.getData();
				String[] datas = data.split(" ");
				
				if (datas[0].equalsIgnoreCase(PACOTE_INTERESSE)) {
					messages.add(new Tuple<Message, Connection>(m,con));	
				}
			}			
		}
		
		if (messages.size() == 0) {
			return null;
		}		
		
		Tuple<Message, Connection> tupleSend = tryMessagesForConnected(messages);	// try to send messages


		if(tupleSend!=null){
				// Obtem a mensagem
			Message m = tupleSend.getKey();
	
			String data = m.getData();
			String[] datas = data.split(" ");
				if(!I.LI.containsKey(m.getId()))
				{
					I.createNewRinterest(m.getId(), m.getData().split(" ")[1]);
					I.setisSendOutInterest(m.getId(), true);
					I.setTimeSendOutInterest(m.getId());
					I.incrSendOutInterests();
				}
	
				if(!I.getisSendOutInterest(m.getId()))
				{
					I.setisSendOutInterest(m.getId(), true);
					I.setTimeSendOutInterest(m.getId());
					I.incrSendOutInterests();
				}

			}
	
		return tupleSend;
	}
	
	
	
	/**
	 * Tries to send all Package Interest to all connected hosts ordered by
	 * their popularity
	 * @return The return value of {@link #tryMessagesForConnected(List)}
	 */
	
	

	public void removemessagesaux(ArrayList<String> messages)
	{
		for (String Id: messages)
		{
			removeFromMessages(Id);
		}
	}
	
	
	/**
	 * Tries to send all Package Content Detail to all connected hosts ordered by
	 * their popularity
	 * @return The return value of {@link #tryMessagesForConnected(List)}
	 */
	/*private Tuple<Message, Connection> trySendPIC() {
		List<Tuple<Message, Connection>> messages = 
			new ArrayList<Tuple<Message, Connection>>(); 
	
		Collection<Message> msgCollection = getMessageCollection();
		
		 for all connected hosts collect all messages that have a higher
		   probability of delivery by the other host 
		for (Connection con : getConnections()) {
			DTNHost other = con.getOtherNode(getHost());
			NamedDataNetworkRouter othRouter = (NamedDataNetworkRouter)other.getRouter();
			
			
			if (othRouter.isTransferring()) {
				continue; // skip hosts that are transferring
			}
			
			for (Message m : msgCollection) {
				if (othRouter.hasMessage(m.getId())) {
					continue; // skip messages that the other one has
				}
				String data = m.getData();
				String[] datas = data.split(" ");
						
				if (datas[0].equalsIgnoreCase(PACOTE_INFORMACAO_CONTEUDO)) {
					// send to other node
					messages.add(new Tuple<Message, Connection>(m,con));	
				}
			}			
		}
		
		if (messages.size() == 0) {
			return null;
		}
		messages = orderByPopularity(messages);
		return tryMessagesForConnected(messages);	// try to send messages
	}
	*/
/*	private List<Tuple<Message, Connection>> orderByPopularity(List<Tuple<Message, Connection>> messages){
		TreeMap<String,Tuple<Message, Connection>> lista = new TreeMap<String,Tuple<Message, Connection>>();
		String data;
		String[] datas;
		PeopleShareInterest newInterest = new PeopleShareInterest();
		int i=0;
		
		for(Tuple<Message, Connection> t : messages){
			data = t.getKey().getData();
			datas = data.split(" ");
			newInterest = li.get(datas[1]);	
			String popularity = String.valueOf(newInterest.getPopularity())+"pop"+i;
			lista.put(popularity, t);
			i++;
		}
		
		List<Tuple<Message, Connection>> lista2 = new ArrayList<Tuple<Message, Connection>>();
		lista2.addAll(lista.descendingMap().values());

		return lista2;
	}
	*/
	@Override
	public NDN_PeopleShareRouter_withLAC replicate() {
		return new NDN_PeopleShareRouter_withLAC(this);
	}
	
	@Override
	public void update() {
		super.update();
		
		if (isTransferring() || !canStartTransfer()) {
			return; // transferring, don't try other connections yet
		}
		
		
		if (isTransferring() || !canStartTransfer()) {
			return; // transferring, don't try other connections yet
		}
		
		// Try first the messages that can be delivered to final recipient
		if (exchangeDeliverableMessages() != null) {
			return; // started a transfer, don't try others (yet)
		}
		
		// then try any/all message to any/all connection
		//this.tryAllMessagesToAllConnections();
		checkSatisfiedInNode();
	
		//this.tryAllMessagesToAllConnections();
		this.trySendPC(); // envia primeiro os PC 
		this.trySendPI();	
	}
}
