/* 
 /* 
 * Copyright 2015 Universidade do Minho, GCOM
 * Modified by Paulo Duarte.
 * Released under GPLv3. See LICENSE.txt for details. 
 */
package routing;


import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import report.NDN_Report;
import report.writeReportToFile;
import routing.NDN.DataReport;
import routing.NDN.InterestReport;
import routing.NDN.PeopleShareInterest;
import core.Connection;
import core.Message;
import core.DTNHost;
import core.Settings;

/**
 * Router
 */
public class NamedDataNetworkRouter extends ActiveRouter {
	
	/** PeopleShare router's setting namespace ({@value})*/ 
	public static final String NDN_NS = "NamedDataNetworkRouter";
	
	public static final String PACOTE_INTERESSE = "pi";
	public static final String PACOTE_CONTEUDO = "pc";
	public static final String PACOTE_INFORMACAO_CONTEUDO = "pic";
	public static final String PACOTE_INTERESSE_ORIGINADO = "pio";
	
	/** lists of interests*/
	protected Map<String, PeopleShareInterest> li;
	/** list of pending interests*/
	protected Map<String, ArrayList<DTNHost>> lpi;
	/** list of pointers to contents*/
	protected Map<String, ArrayList<DTNHost>> lpc;
	/** list of contents*/
	public Map<String, String> lc;
	
	protected Map<String, ArrayList<DTNHost>> lei;
	
	private boolean hop = false;

	boolean LC = false; //LoadContent
	public static final String CONTENT = "content";
	public static final String HOSTS = "nrofHosts";
	public String  contentList = "";
	private int nrofHosts;
	private int id = 0;
	
	DataReport R;
	InterestReport I;
	writeReportToFile wtf;
	
	public NamedDataNetworkRouter(Settings s) {
		super(s);		
		Settings peopleShareSettings = new Settings(NDN_NS);
		contentList = peopleShareSettings.getSetting(CONTENT);
		nrofHosts = peopleShareSettings.getInt(HOSTS);
		initRouter();
	}
	
	protected NamedDataNetworkRouter(NamedDataNetworkRouter r) {
		super(r);
		this.contentList = r.contentList ;
		this.nrofHosts = r.nrofHosts;
		initRouter();
	}
	
	public void initRouter(){
		this.li = new HashMap<String, PeopleShareInterest>();
		this.lpi = new HashMap<String, ArrayList<DTNHost>>();
		this.lpc = new HashMap<String, ArrayList<DTNHost>>();
		this.lc = new HashMap<String, String>();
		this.lei = new HashMap<String, ArrayList<DTNHost>>();

		if(contentList.equalsIgnoreCase("lista_de_interesses.txt")){
			loadRndContent();
		}
		
		R = DataReport.getInstance();
		I = InterestReport.getInstance();
		wtf = writeReportToFile.getInstance();
	}
	
	public Map<String, String> getLc()
	{
		return this.lc;
	}
	
	
	private void loadContent(){		
		String line;
		String[] contents, path;
		try{
			File arquivo = new File(contentList);
			FileReader fr = new FileReader(arquivo);
			BufferedReader br = new BufferedReader( fr );
			int n = 0;
			while(br.ready()){		
				line = br.readLine();
				contents = line.split(" "); //contents[0] - host ; contents[1] - content
				path = contents[1].split("/"); 
				if(this.getHost().getAddress() == Integer.valueOf(contents[0])){
					addInterest("I" + n, contents[1], 0, false, contents[1]);
					n++;	
					if(NDN_Report.list_of_publishers.containsValue(path[0])){
						continue;
					}
					else{
						NDN_Report.list_of_publishers.put(Integer.valueOf(contents[0]), path[0]);
					}
				}
			}
			br.close();
			fr.close();
		}
		catch(Exception e){
			System.out.println("Erro " + e + " na leitura de " + contentList);
		}
	}

	private void loadRndContent(){		
		String line;
		String[] path;
		boolean temp = false;
		int rndHost = getRandomHost();
		
		try{
			File arquivo = new File(contentList);
			FileReader fr = new FileReader(arquivo);
			BufferedReader br = new BufferedReader( fr );
			while(br.ready()){		
				line = br.readLine();
				path = line.split("/");
				if(NDN_Report.list_of_publishers.containsValue(path[0])){
					continue;
				}
				else{
					do{
						if(NDN_Report.list_of_publishers.containsKey(rndHost)){
							rndHost = getRandomHost();
							
					}
						else{
							temp = true;
							NDN_Report.list_of_publishers.put(rndHost, path[0]);
							
						}
					}while(temp == false);
				}
			}
			br.close();
			fr.close();
		}
		catch(Exception e){
			System.out.println("Erro " + e + " na leitura de " + contentList);
		}
	}
	
	private void uploadContent(){
		String line;
		String[] path;
		try{
			File arquivo = new File(contentList);
			FileReader fr = new FileReader(arquivo);
			BufferedReader br = new BufferedReader( fr );
			int n = 0;
			if(NDN_Report.list_of_publishers.containsKey(this.getHost().getAddress())){
				String domain = NDN_Report.list_of_publishers.get(this.getHost().getAddress());
				while(br.ready()){		
					line = br.readLine();
					path = line.split("/"); 					
					if(domain.equalsIgnoreCase(path[0])){
						addInterest("I" + n, line, 0, false, line);
						n++;
						printLc();
					}
				}
				br.close();
				fr.close();
			}

		}
		catch(Exception e){
			System.out.println("Erro " + e + " na leitura de " + contentList);
		}
	}

	public int getRandomHost(){
		Random r = new Random();
		int n = r.nextInt(nrofHosts + 1);
		return n;
	}
	
	public void addInterest(String id, String name, int nrOfRequests, boolean isSubs, String content){
		PeopleShareInterest newInterest = new PeopleShareInterest();
		ArrayList<DTNHost> novaLista = new ArrayList<DTNHost>();
		if(!li.containsKey(name)){
			newInterest = new PeopleShareInterest(id,name,nrOfRequests,isSubs,0);
			li.put(name, newInterest);
		}
		String idInterest = li.get(name).getId();
		if(!lpc.containsKey(idInterest)){
			lpc.put(idInterest, novaLista);
		}
		if(!(content == null)){
			lc.put(name, content);
		}
	}
	
	public String getNewId(){
		this.id++;
		String newId = "I" +id;
		if(id == 82) id=0;
		return newId;
	}
		
	public void printLi(){
		System.out.println("\n*** LISTA DE INTERESSES de " + this.getHost().toString() + " ***\n");
		for(PeopleShareInterest inter : this.li.values()){
			System.out.println( "Ass: " + inter.getIsSubscriber() + 
					"\t\tNpedidos: " + inter.getNrOfRequests() +
					"\t\tPopularity: " +inter.getPopularity()+
					"\t\tID: " + inter.getId() +
					"\t\tNome: " + inter.getName());
		}
	}
	
	public void printLpi(){
		String name;
		System.out.println("\n*** LISTA DE INTERESSES PENDENTES de " + this.getHost().toString() + " ***\n");
		for(String idInt : this.lpi.keySet()){
			for(DTNHost host : this.lpi.get(idInt)){
				name = getNameById(idInt);
				System.out.println("Host: " +host.toString()+ "\tInteresse: " +name);
			}
		}
	}
	
	public void printLpc(){
		String name;
		System.out.println("\n*** LISTA DE APONTADORES PARA CONTEDOS de " + this.getHost().toString() + " ***\n");
		for(String idInt : this.lpc.keySet()){
			for(DTNHost host : this.lpc.get(idInt)){
				name = getNameById(idInt);
				System.out.println("Host: " +host.toString()+ "\tInteresse: " +name);
			}
		}
	}
	
	public void printLc(){
		System.out.println("\n*** LISTA DE CONTEDOS de " + this.getHost().toString() + " ***\n");
		for(String content : this.lc.values()){
			System.out.println("Contedo: " + content);
		}
	}
	
	private String getNameById(String id){
		String path = null;
		for(PeopleShareInterest i: li.values()){
			if(i.getId() == id){
				path = i.getName();
			}
		}
		return path;
	}
	
	/**
	 * Generates a (random) message size
	 * @return message size
	 */
	protected int drawMessageSize() {
		/*Random r = new Random();
		int sizeDiff = NDN_Report.sizeRangeLow == NDN_Report.sizeRangeHigh ? 0 : 
			r.nextInt(NDN_Report.sizeRangeHigh - NDN_Report.sizeRangeLow);
		return NDN_Report.sizeRangeLow + sizeDiff;*/
		
		Random r = new Random();
		int sizeRangeHigh = 400000; // avg webpage
		int sizeRangeLow = 200000; //avg musica
		//int sizeDiff = NDN_Report.sizeRangeLow == NDN_Report.sizeRangeHigh ? 0 : 
		//	r.nextInt(NDN_Report.sizeRangeHigh - NDN_Report.sizeRangeLow);
		int sizeDiff = sizeRangeLow == sizeRangeHigh ? 0 : 
					r.nextInt(sizeRangeHigh - sizeRangeLow);
		return sizeRangeLow + sizeDiff;
	}
	
	public int getMessageLenght(Message m, String data){
		int lenght = m.getFrom().toString().length() + data.length();
		return lenght;
	}
	
	/**
	 * Called when a connection's state changes. This version doesn't do 
	 * anything but subclasses may want to override this.
	 */
	@Override
	public void changedConnection(Connection con) { }
	
	@Override
	public Message messageTransferred(String id, DTNHost from) {
		Message m = super.messageTransferred(id, from); 
		return m;
	}
	
	@Override
	public void update() {
		super.update();
		if(!LC){
			LC=true;
			if(contentList.equalsIgnoreCase("lista_de_interesses.txt")){
				uploadContent();
			}
			else
				loadContent();			
		}
		Collection<Message> msgCollection = getMessageCollection();
		Collection<Message> temp = new ArrayList<Message>();
		temp.addAll(msgCollection);
		for (Message m : temp) {
			DTNHost other = m.getFrom();
			NamedDataNetworkRouter othRouter = (NamedDataNetworkRouter)other.getRouter();
			PeopleShareInterest newInterest = new PeopleShareInterest();
			String data = m.getData(); 
			//datas[0] - type ; datas[1] - name/path ; datas[2] - content;
			String[] datas = data.split(" ");
			String[] path = datas[1].split("/");
			
			if(datas[0].equalsIgnoreCase(PACOTE_INTERESSE)){ 
				//Se no existir na LI, adicionar novo
				if(datas.length == 3){
					if(datas[2].equalsIgnoreCase("pio")){
						removeFromMessages(m.getId());
						int lenght = getMessageLenght(m, PACOTE_INTERESSE+" "+ datas[1]);
						if(hop){
							m = new Message(m.getFrom(), m.getFrom(),
								m.getId(),lenght, PACOTE_INTERESSE+" "+m.getFrom()+"#0|"+ datas[1]);		
						}
						else{
							
							m = new Message(m.getFrom(), m.getFrom(),
									m.getId(),lenght, PACOTE_INTERESSE+" "+ datas[1]);
						}
							
						if(othRouter.li.containsKey(datas[1])){
							newInterest=othRouter.li.get(datas[1]);
							if(newInterest.getIsSubscriber()){
								NDN_Report.final_destiny2++; // aqui quer dizer que o interesse foi satisfeito localmente
								
							}
							else{
								if(createNewMessage(m)){
									NDN_Report.interests_originated++;
									if(!othRouter.li.containsKey(datas[1])){ 
										othRouter.addInterest(m.getId(),datas[1],0,true,null);
									}
									else{
										newInterest = othRouter.li.get(datas[1]);
										newInterest.setIsSubscriber(true);
										
										othRouter.li.put(datas[1], newInterest);
										if(!I.LI.containsKey(m.getId()))
										{
											I.createNewRinterest(m.getId(), datas[1]);
										}
									}
									
									if(!I.LI.containsKey(m.getId()))
									{
										I.createNewRinterest(m.getId(), datas[1]);
									}
									
								}
							}
						}
						else{
							if(createNewMessage(m)){
								NDN_Report.interests_originated++;
								othRouter.addInterest(m.getId(),datas[1],0,true,null);
								if(!I.LI.containsKey(m.getId()))
								{
									I.createNewRinterest(m.getId(), datas[1]);
								}
							}
							else
								System.out.println("No criei msgs");
						}
					}
				}
			}
		}
		wtf.writeNumberInterestsSastified(I, R);
		wtf.AvgDelaywithCI(I);
		wtf.AvgDelaywithCIAlternative(I);
	}
	
	@Override
	protected void dropExpiredMessages() {
		
		Message[] messages = getMessageCollection().toArray(new Message[0]);
		
		//List<Tuple<Message, Connection>> outgoingMessages = 
		//		new ArrayList<Tuple<Message, Connection>>(); 
		
		List<Message> msgs = new ArrayList<Message>();
		for (int i=0; i<messages.length; i++) {
			int ttl = messages[i].getTtl(); 
			if (ttl <= 0) {
				// trysendInterest com flag especial 
				if(messages[i].getData().split(" ")[0].equalsIgnoreCase("pi")) 
				{
						deleteMessage(messages[i].getId(), true);
						checklpi(messages[i].getData().split(" ")[1]);
				}
				
				if(messages[i].getData().split(" ")[0].equalsIgnoreCase("pc")) 
				{
					deleteMessage(messages[i].getId(), true);
					checkLC(messages[i].getData().split(" ")[1]);
				}
			}
		}
		
		
	}
	
	
	public void checkLC(String name)
	{
		if(this.lc.containsKey(name))
		{
			for (Message m: getMessageCollection())
			{
				String datas [] = m.getData().split(" ");
				if(datas[0].equalsIgnoreCase("pc"))
				{
					if(datas[1].equalsIgnoreCase(name)) return;
				}
			}
			if(this.deliveredMessages.containsKey(name))
			{
				return;
			}
			this.lc.remove(name);
		}
	}
	
	
	public void checklpi(String name)
	{
		String idInt = null;
		
		if(li.containsKey(name))  idInt = li.get(name).getId(); 
		
		if(this.lpi.containsKey(name))
		{
			for (Message m: getMessageCollection())
			{
				String datas [] = m.getData().split(" ");
				if(datas[0].equalsIgnoreCase("pi"))
				{
					if(datas[1].equalsIgnoreCase(name)) return;
				}
			}
			this.lpi.remove(name);
		}
	}
	
	@Override
	public NamedDataNetworkRouter replicate() {
		return new NamedDataNetworkRouter(this);
	}
}

