/* 
 /* 
 * Copyright 2015 Universidade do Minho, GCOM
 * Modified by Paulo Duarte.
 * Released under GPLv3. See LICENSE.txt for details. 
 */

package routing;

	//onData
	// onInterest
	// onFIBDigest
	// TrySendOutInterest
	// TrySendOutData
	// TrySendOutFIBDigest
	// receivedDuplicatedInterest
	// updateDeliveryPredsFor
	// updateTransitivyPredsFor
	// SatisfyPendingInterest

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.Map.Entry;

import routing.util.RoutingInfo;
import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;
import core.SimClock;
import core.Tuple;

public class PIFP extends NDNRouter{
	
	public Map<String, ArrayList<DTNHost>> inPC;

	public PIFP(Settings s) {	
		super(s);
		// TODO Auto-generated constructor stub
	}
	
	public PIFP(PIFP r)
	{
		super(r);
		this.inPC = new HashMap<String, ArrayList<DTNHost>>();
	}


	@Override
	public void changedConnection(Connection con) {
		
		if (con.isUp()) 
		{
			
			DTNHost otherHost = con.getOtherNode(getHost());
			updateDeliveryPredFor(otherHost); 
			updateTransitivePreds(otherHost);
		}
		
	}
	

	@Override
	public Message messageTransferred(String id, DTNHost from) {	
		/** 
		 * datas[0] = type
		 * datas[1] = name
		 * datas[2] = content
		 */
		
		Message m = super.messageTransferred(id, from); 
		String data = m.getData();
		String[] datas = data.split(" ");

		//When receiving PI
		if(datas[0].equalsIgnoreCase(INTEREST_PACKAGE)){		
			if(!(m.getFrom() == this.getHost())) onInterest(m, datas[1], from);
			if((m.getFrom() == this.getHost())) this.deleteMessage(m.getId(), true);
		}
		
		//When receiving PC
		if(datas[0].equalsIgnoreCase(CONTENT_PACKAGE)){
			if(!(m.getFrom() == this.getHost())) onData(m, datas[1], from);	
			if((m.getFrom() == this.getHost())) this.deleteMessage(m.getId(), true);
		}
		return m;
	}
	
	/**
	 * Remove set of messages
	 * @param ArrayList Messages
	 */
	private void removemessage(ArrayList<String> messages)
	{			
		for (String Id : messages)
		{
			this.deleteMessage(Id, true);
		}
	}
	
	
	
	/**
	 * Remove all messages of interest already satisfied.
	 * If there are messages generated by the node itself increase
	 * messages locally satisfied
	 */
	public void checkSatisfiedMessages()
	{
		Collection<Message> msgCollection = this.getMessageCollection();
		ArrayList<String> idtoRemove = new ArrayList<String>() ;

		for (Message m : msgCollection) 
		{
			String data = m.getData();
			String[] datas = data.split(" ");

			if (datas[0].equalsIgnoreCase(INTEREST_PACKAGE)) 
			{
				if(this.lc.containsKey(datas[1]))
				{
					deletePITEntrie(datas[1]);
					if(m.getFrom() == this.getHost())
					{
						I.numberInterestsSastifiedinNode();
							
						if(!idtoRemove.contains(m.getId()))
						{
							idtoRemove.add(m.getId());	
						}
					}else
					{
						if(!idtoRemove.contains(m.getId()))
						{
							idtoRemove.add(m.getId());	
						}
					}
				}
			}
		}
		
		removemessage(idtoRemove);
	}
	
	
	/**
	 * Updates delivery predictions for a host.
	 * <CODE>P(a,content_b) = P(a,content_b)_old + (1 - P(a,content_b)_old) * P_INIT</CODE>
	 * @param host The host we just met
	 */
	private void updateDeliveryPredFor(DTNHost to) {
		
		Double oldValue = 0.0;	
		MessageRouter otherRouter = to.getRouter();
		Map<String, Integer> otherContents = ((PIFP)otherRouter).lc; // get list of information content from the other node
		
		for (String name : otherContents.keySet())
		{
			//check if the content already exists in the CS itself
			if (!(this.bloomLC.contains(name)) && !(checkifExistesDeliveredmessages(name)))
			{
				oldValue = getPredFor(name);
				if (oldValue == null) oldValue = 0.0;
			 
				double newValue = oldValue + (1 - oldValue) * P_INIT; 

				this.FIB.put(name, newValue); // add new entry in the FIB

				if(this.bloomPIT.contains(name)) 
				{
					for (Message m : this.getMessageCollection()) 
					{
						String datas [] = m.getData().split(" ");
					
						if (to.getRouter().hasMessage(m.getId())) 
						{
							continue; // skip messages that the other one has
						}
						 
						if (datas[0].equalsIgnoreCase("pi"))
						{
							if (datas[1].equalsIgnoreCase(name))
							{
								if( sendMessage(m.getId(), to) <1) 
								{ 	
									if(I.LI.containsKey(m.getId()))
									{
										if(!(I.getisSendOutInterest(m.getId())))
										{
											I.setTimeSendOutInterest(m.getId());
											I.setisSendOutInterest(m.getId(), true);								 
											I.incrSendOutInterests();
										} 
									}
								}		 
							}
						}
					}
				
				}	
				
			}
		}
		return;
	}
	
	

	/**
	 * Updates transitive (A->B->Content_C) delivery predictions.
	 * <CODE>P(a,content_c) = P(a,content_c)_old + (1 - P(a,content_c)_old) * P(a,b) * P(b,content_c) * BETA
	 * </CODE>
	 * @param host The B host who we just met
	 */
	private void updateTransitivePreds(DTNHost host) {
		MessageRouter otherRouter = host.getRouter();
		double pa_c_old =0; //P(a,c)_old
		double  pa_b = 0.75; // P(a,b)

		// get predictabilities the other host
		Map<String, Double> otherPreds = ((PIFP)otherRouter).getDeliveryPreds();
		
		
		Map<String, Integer> otherContent = ((PIFP)otherRouter).lc; // get names of contents of the other host
		
		for (String e : otherPreds.keySet()) 
		{
			double pb_c = 0;
			// verifies that does not contain any existing content in CS
			if (this.bloomLC.contains(e) || checkifExistesDeliveredmessages(e) || otherContent.containsKey(e)) 
			{
				continue;
			}
			
			pb_c = otherPreds.get(e); // get the predictability of node b reach the content c
			pa_c_old = getPredFor(e); // P(a,c)_old - predictability of node a reach the content c
	
			double pNew = pa_c_old + ( 1 - pa_c_old) * pa_b * pb_c * beta; //p(a,c) = P(a,c)old + (1-P(a,c)old).P(a,b).P(b,c)xBeta
			this.FIB.put(e, pNew);
			
			if(!this.bloomFIB.contains(e)) 
			{
				this.bloomFIB.add(e);
			}
			
		}
	}

	
	
	/**
	 * Returns the current prediction (P) value for a host or 0 if entry for
	 * the host doesn't exist.
	 * @param host The host to look the P for
	 * @return the current P value
	 */
	public double getPredFor(String name) {
		ageDeliveryPreds(); // make sure preds are updated before getting
		if (this.FIB.containsKey(name)) {
			return this.FIB.get(name);
		}else
		{
			return 0;
		}
	}
	


	/**
	 * Ages all entries in the delivery predictions.
	 * <CODE>P(a,b) = P(a,b)_old * (GAMMA ^ k)</CODE>, where k is number of
	 * time units that have elapsed since the last time the metric was aged.
	 * @see #SECONDS_IN_UNIT_S
	 */
	private void ageDeliveryPreds() 
	{
		Double value;
		double timeDiff = (SimClock.getTime() - this.lastAgeUpdate) / secondsInTimeUnit;

		if (timeDiff == 0){
			this.lastAgeUpdate = SimClock.getTime();
			return;
		}
		
		double mult = Math.pow(GAMMA, timeDiff);
		
		if (this.FIB.isEmpty()){
				return;
		}
		
		for (String e : this.FIB.keySet()){
			value = this.FIB.get(e);
			this.FIB.put(e, mult*value);  
		}
		
		this.lastAgeUpdate = SimClock.getTime();
	}
	
	/**
	 * Returns a map of this router's delivery predictions
	 * @return a map of this router's delivery predictions
	 */
	private HashMap<String, Double> getDeliveryPreds() 
	{
		ageDeliveryPreds(); // make sure the aging is done
		return this.FIB;
	}
	
	
	
	/**
	 * Called when it receives data and possesses pending interest messages.
	 * A message of interest is eliminated, created by the node itself.
	 * @param name is the name of content;
	 * @return "removed", if a message of interest is removed
	 */	
	public String removeInterest(String name)
	{
		String data="";
		String datas[];
		Collection<Message> msgCollection = getMessageCollection();	
		
		for (Message m : msgCollection) 
		{
			data = m.getData();
			datas = data.split(" ");
			
			if(datas[1].equalsIgnoreCase(name))
			{
				if(datas[0].equalsIgnoreCase(INTEREST_PACKAGE))
				{
					if((m.getFrom() == this.getHost()))
					{
						this.deleteMessage(m.getId(), true);
						return null;
					}
				}
			}
		}
		return null;	
	}
	
	
	/**
	 * Processing the data packets received from the network. 
	 * If the node is subscriber data is added to the repository. 
	 * If not are just kept in the CS temporarily.
	 * @param m is message
	 * @param name is the name of content
	 * @param from is the sender
	 */
	public void onData(Message m, String name, DTNHost from)
	{	
		ArrayList<DTNHost> allHosts = new ArrayList<DTNHost>();
		boolean isubscriber = false;
		
		if (bloomPIT.contains(name))
		{
			if (this.PIT.containsKey(name))
			{
				allHosts = this.PIT.get(name).getAllHosts(); // get all nodes interested in the data
				
				
				if(allHosts.contains(this.getHost())) // checks if is subscriber or not
				{
					String idm[] = m.getId().split("_");
					createCSEntrie(name, m, 2); // create Content Store entry
					deliveredMessages.put(m.getId(), m); // add message to repository 
					
					I.createPITSast(this.PIT.get(name).timeCreated, SimClock.getTime()); // reports
					deletePITEntrie(name);
					R.incrNrHops(m); //reports
						
					// 	If no other interfaces, removes interest
					if(allHosts.size() <= 1) removeFromMessages(m.getId());
							
					
					removeInterest(name); 
							
					// Reports ------------------
					if (I.LI.containsKey(idm[1])){
						I.settimeSastifiedInterest(idm[1]);
						I.setisSatisfiedInterest(idm[1], true);
					}
	
					if (R.PC.containsKey(m.getId()))
					{	
						R.timeDelivered(m.getId());
						R.setDeliveredData(m.getId(),  true);
						
					}
						
					isubscriber = true;
					R.nrDeliveredData(); 
					I.incrSastifiedInterests(); 	
					//----------------------------
				}
				else 
				{
					if(this.lei.containsKey(name))
					{
							if(!this.lei.get(name).contains(host)) this.lei.get(name).add(host);
					}
						else 
						{
							ArrayList<DTNHost> aux = new ArrayList<DTNHost>();
							aux.add(host);
							this.lei.put(name, aux);
						}	
				}
				
				if(isubscriber!=true)
				{
					//Add CS entry
					if(!this.lc.containsKey(name))
					{
						createCSEntrie(name, m, 3); 
					}
					
					deletePITEntrie(name);//Delete PIT entry
					removeInterest(name);//Delete message of interest
				}
			}
		}
	}
	
	
	/**
	 * Adds new face to send data message. The list is to assist the PIT. 
	 * @param c is message of content
	 * @param name is the name of content
	 * @param othRouter is the face to send 
	 */
	public void sendDataMessage(Message c, String name, NDNRouter othRouter)
	{			
		 if(this.lei.containsKey(name))
		 {
			 if(!this.lei.get(name).contains(othRouter.getHost()))
				 this.lei.get(name).add(othRouter.getHost());
		 }
		 else {
			 ArrayList<DTNHost> aux = new ArrayList<DTNHost>();
			 aux.add(othRouter.getHost());
			 this.lei.put(name, aux);
		 }
		 trySendData();		
	}
	

	/**
	 * Reception of a interst packet that is treated. this time the packet is analyzed
	 * and the conditions for acceptance of new interest are checked. 
	 * If any content that satisfies the interest is returned immediately, 
	 * otherwise the PIT is added a new entry with the conditions.
	 * @param m is message
	 * @param name is the name of content
	 * @param from is the sender
	 */
	public void onInterest(Message m, String name, DTNHost from)
	{	
		
		String[] data = name.split("/");
		int size = data.length;
	
		NDNRouter othRouter = (NDNRouter)from.getRouter();
		Message c = null;
		
		if (this.bloomLC.contains(name)) // if there is content in CS
		{
			// reports
			if(I.existsIdMessage(m.getId()) == false) I.incrNrHops(m); // increase number of hops
			this.nrinterestsreceived = this.nrinterestsreceived + 1; // increase number of interests received
		
			
			if(!this.hasMessage(RESPONSE_PREFIX+m.getId()) && !this.deliveredMessages.containsKey(RESPONSE_PREFIX+m.getId())) //if no message exists in the buffer or message in repository
			{
				c = new Message(this.getHost(), othRouter.getHost(), RESPONSE_PREFIX+m.getId(), drawMessageSize(), CONTENT_PACKAGE+ " " + name + " " +name);
				if(createNewMessage(c))
				{
					c.setTtl(500);
					sendDataMessage(c, name,  othRouter);//send PC
				}	 
			}
			else {
				if(!this.hasMessage(RESPONSE_PREFIX+m.getId())) 
				{
					if (this.deliveredMessages.containsKey(RESPONSE_PREFIX+m.getId()))
					{
						Message msg = this.deliveredMessages.get(RESPONSE_PREFIX+m.getId());
						msg.setTo(othRouter.getHost());
						addToMessages(msg, false);
						sendDataMessage(msg, name, othRouter);
					}
				}
				else //there ready to send message
				{
					Collection<Message> msgCollection = getMessageCollection();

					for(Message d : msgCollection )
					{
						if (d.getId().equalsIgnoreCase(RESPONSE_PREFIX+m.getId()))
						{
							 d.setTo(othRouter.getHost());
							 sendDataMessage(d,  name, othRouter);
							 break;
						}
					}	
				}
			}
			if(I.LI.containsKey(m.getId())) 
			{
				I.settimeDelivered(m.getId());
			}
		
		}
		else if(bloomFIB.contains(name)) // if there is knowledge of the content
		{
			this.nrinterestsreceived = this.nrinterestsreceived + 1;
			insertPI(m, false, from); //insert interest in PIT and subscriber is false
			if (!bloomPIT.contains(name)) 
			{
				bloomPIT.add(name);
			}
		}
		else
		{ 	// if you do not have bloom FIB entry for the interest, discard the message	
			this.toremove.put(m.getId(), SimClock.getTime());
		}
		
	}
	
	
	
	/**
	 * Try send Dato to other router
	 * @return Tuple<Message, Connection>
	 */
	private Tuple<Message, Connection> trySendData()
	{
		List<Tuple<Message, Connection>> outgoingMessages = new ArrayList<Tuple<Message, Connection>>(); 
		Collection<Message> msgCollection = getMessageCollection();
			
		/* for all connected hosts collect all messages that have a higher probability of delivery by the other host */
		
		for (Connection con : getConnections()) 
		{
			DTNHost other = con.getOtherNode(getHost());
			NDNRouter othRouter = (NDNRouter)other.getRouter();
				
			if (othRouter.isTransferring())
			{
				continue; // skip hosts that are transferring
			}

			for (Message m : msgCollection) 
			{
				if (othRouter.hasMessage(m.getId())) {
					continue; // skip messages that the other one has
				}
						
				String data = m.getData();
				String[] datas = data.split(" ");
				String name = datas[1];

				if(datas[0].equalsIgnoreCase(CONTENT_PACKAGE)) 
				{
					
					if(this.lei.containsKey(name)) 
					{
						if(this.lei.get(name).contains(con.getOtherNode(getHost())))
						{
							outgoingMessages.add(new Tuple<Message, Connection>(m,con));
						}
					}
				}		
			} 
			
	
			if (outgoingMessages.size() == 0) {
				return null;
			}
				
			Tuple<Message, Connection> tupleSend = tryMessagesForConnected(outgoingMessages);	// try to send messages
		
			if(tupleSend!=null)
			{		
				Message send = tupleSend.getKey(); 
				Connection c = tupleSend.getValue();
				DTNHost h = c.getOtherNode(getHost());
				String[] dat = send.getData().split(" ");
					
				if (!(R.PC.containsKey(send.getId()))) R.createNewPC(send.getId(), (send.getData().split(" "))[1], true);
				
				R.incrSendOutData();
				this.nrdatasent = this.nrdatasent +1;
						
				if(this.lei.containsKey(dat[1]))
				{
					if(this.lei.get(dat[1]).size()<=1) 
					{
						this.lei.remove(dat[1]);
					}
					else  this.lei.get(dat[1]).remove(h); // se no for o unico remove apenas o n para quem enviar 
				}		
			}
		}
		return null;
	}

	/**
	 * Remove set of messages
	 * @param ArrayList Messages
	 */
	public void removemessages(ArrayList<String> messages)
	{			
		for (String Id : messages)
		{
			removeFromMessages(Id);
		}
	}
	

	/**
	 * Try send Interest to other router
	 * @return Tuple<Message, Connection>
	 */
	private Tuple<Message, Connection> trySendInterest() 
	{
			
		ArrayList<String> idtoRemove = new ArrayList<String>() ;
					
		List<Tuple<Message, Connection>> outgoingMessages = 
				new ArrayList<Tuple<Message, Connection>>(); 
			
		List<Tuple<Message, Connection>> out = 
					new ArrayList<Tuple<Message, Connection>>(); 
		
		Collection<Message> msgCollection = getMessageCollection();
		
		/* for all connected hosts collect all messages that have a higher
			   probability of delivery by the other host */
			
		checkSatisfiedMessages();
		
		for (Connection con : getConnections()) 
		{
			DTNHost other = con.getOtherNode(getHost());
			NDNRouter othRouter = (NDNRouter)other.getRouter();
					
			if (othRouter.isTransferring()) {
				continue; // skip hosts that are transferring
			}
				
			for (Message m : msgCollection) 
			{
				if (othRouter.hasMessage(m.getId())) {	
					continue; // skip messages that the other one has
				}

				String data = m.getData();
				String[] datas = data.split(" ");	

				if(this.bloomPIT.contains(datas[1]))
					{
						if(othRouter.bloomLC.contains(datas[1])) //if the other host has data
						{
							out.clear();
							out.add(new Tuple<Message, Connection>(m,con));				
							Tuple<Message, Connection> res = tryMessagesForConnected(out);

							if(res!=null)
							{			
								if(!I.getisSendOutInterest(res.getKey().getId()))
								{
									I.setisSendOutInterest(res.getKey().getId(), true);
									I.setTimeSendOutInterest(res.getKey().getId());
									I.incrSendOutInterests();
								}
							}
						}	
						else  if (othRouter.bloomFIB.contains(datas[1])) // If the other node has in its FIB and prev is> that this node prev
						{
							if (comparePrevs(this, othRouter, datas[1])>=1)
							{	
								outgoingMessages.add(new Tuple<Message, Connection>(m,con));
							}
						}		
					}
				}
			
		
		
			if (idtoRemove.size() != 0)
			{
				removemessages(idtoRemove);
			}
			
			if (outgoingMessages.size() == 0) {
				return null;
			}		
				
			//sort messages in descending order with respect to the probabilities of delivery
			outgoingMessages = orderByPrev(outgoingMessages);
			
			// Try to send messages to the corresponding connections 
			// Returns the message and connection to which it was sent
			Tuple<Message, Connection> tupleSend = tryMessagesForConnected(outgoingMessages);
				
			if(tupleSend!=null)
			{
				// Gets the message
				Message send = tupleSend.getKey(); 
				
				
				//Reports
				// Set the time at which the message was sent
				if(!I.getisSendOutInterest(send.getId()))
				{
					I.setisSendOutInterest(send.getId(), true);
					I.setTimeSendOutInterest(send.getId());
					I.incrSendOutInterests();
				}
			}
		}
		return null;
			
	}

		

	/**
	 * Order Set of Messages By Probability of reach the producer
	 * @param List<Tuple<Message, Connection>>  Messages
	 * @return  List<Tuple<Message, Connection>> Messages ordered by Probability
	*/
	private List<Tuple<Message, Connection>> orderByPrev(List<Tuple<Message, Connection>> messages)
	{
		TreeMap<String,Tuple<Message, Connection>> lista = new TreeMap<String,Tuple<Message, Connection>>();
		String data;
		String[] datas;
		int i=0;
			
		for(Tuple<Message, Connection> t : messages)
		{
			data = t.getKey().getData();
			datas = data.split(" ");
			double prev;
				
			if (this.FIB.containsKey(datas[1]))
			{
				prev = this.FIB.get(datas[1]);
			}
			else prev=1;
				
			String pred = String.valueOf(prev+"pop"+i);
			lista.put(pred, t);
			i++;
		}
			
		List<Tuple<Message, Connection>> lista2 = new ArrayList<Tuple<Message, Connection>>();
		lista2.addAll(lista.descendingMap().values());

		return lista2;
			
	}
			

	
	@Override
	public RoutingInfo getRoutingInfo() {
		ageDeliveryPreds();
		RoutingInfo top = super.getRoutingInfo();
		RoutingInfo ri = new RoutingInfo(FIB.size() + 
				" delivery prediction(s)");
		
		for (Entry<String, Double> e : FIB.entrySet()) {
			String name = e.getKey();
			Double value = e.getValue();
			
			ri.addMoreInfo(new RoutingInfo(String.format("%s : %.6f", 
					name, value)));
		}
		
		top.addMoreInfo(ri);
		return top;
	}
	
	
	
	/**
	 * Called in sending messages of interest to determine if the other node has more chances to get to producer
	 * @param r1 this router
	 * @param r2 other router
	 * @param name of the interest
	 * @return -1 if the probability of this router is greater than the other
	 */
	public int comparePrevs(NDNRouter r1, NDNRouter r2, String name)
	{
		double p1; // this router
		double p2; // other router
		
		double lastProb= this.PIT.get(name).getLastProb(); // get last probability P(b,content)
		
		if (r1.FIB.containsKey(name)) p1 = r1.FIB.get(name);
		else p1 = 0;
		
		if (r2.FIB.containsKey(name)) p2 = r2.FIB.get(name);
		else p2=0;
		
		// bigger probability should come first
		if (p2-p1 == 0) 
		{			
			if (p2-lastProb > 0)
			{
				this.PIT.get(name).setLastProb(p2);
				return 1;
			}
			else return -1;
		}
		else if (p2-p1 < 0) {
			return -1;
		}
		else {
			if (p2-lastProb > 0)
			{
				this.PIT.get(name).setLastProb(p2);
				return 1;
			}
			else return -1;
		}		
		
	}


	
	@Override
	public PIFP replicate() {
		return new PIFP(this);
	}

	
	
	@Override
	public void update() {
		super.update();
		if (isTransferring() || !canStartTransfer()) {
			return; // can't start a new transfer
		}
		
		// communication is driven by sending messages of interest 
		this.trySendInterest(); // try send Interest first 
		this.trySendData();	// try send data 
		
	
	}

}
