/* 
 /* 
 * Copyright 2015 Universidade do Minho, GCOM
 * Created by Paulo Duarte.
 * Released under GPLv3. See LICENSE.txt for details. 
 */

package routing;

	//onData
	// onInterest
	// onFIBDigest
	// TrySendOutInterest
	// TrySendOutData
	// TrySendOutFIBDigest
	// receivedDuplicatedInterest
	// updateDeliveryPredsFor
	// updateTransitivyPredsFor
	// SatisfyPendingInterest
	
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.Map.Entry;

import routing.PeopleShare.InterestList;
import routing.util.RoutingInfo;
import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;
import core.SimClock;
import core.Tuple;

public class PIFP_Proactive extends NDNRouter{
	
	private double lastupdate;
	
	public PIFP_Proactive(Settings s) 
	{	
		super(s);
		// TODO Auto-generated constructor stub
	}
	
	public PIFP_Proactive(PIFP_Proactive r)
	{	
		super(r);	
		lastupdate = 0;
	}


	@Override
	public void changedConnection(Connection con) 
	{
		if (con.isUp()) 
		{
			
			DTNHost otherHost = con.getOtherNode(getHost());
			updateDeliveryPredFor(otherHost); 
			updateTransitivePreds(otherHost);	
		}
	}
	
	

	@Override
	public Message messageTransferred(String id, DTNHost from) {
		Message m = super.messageTransferred(id, from); 
		String data = m.getData();
		
		/** 
		 * datas[0] = type
		 * datas[1] = name
		 * datas[2] = content
		 */
		
		String[] datas = data.split(" ");

		//When receiving PI
		if(datas[0].equalsIgnoreCase(INTEREST_PACKAGE)){	
		
			if(!(m.getFrom() == this.getHost())) onInterest(m, datas[1], from);
			
			if((m.getFrom() == this.getHost()))
			{
				this.deleteMessage(m.getId(), true);
			}
			
		}
		
		//When receiving PC
		if(datas[0].equalsIgnoreCase(CONTENT_PACKAGE)){
			
			if(!(m.getFrom() == this.getHost())) onData(m, datas[1], from);
			
			if((m.getFrom() == this.getHost()))
			{
				this.deleteMessage(m.getId(), true);
			}
			
		}

		return m;
	}
	
	/**
	 * Calculate popularity 
	 * Popularity = dataSent/intReceived
	 **/
	public void calcpopularity()
	{
		if(this.nrdatasent>0 && this.nrinterestsreceived>0)
		{
			double datasent = (double)this.nrdatasent;
			double intreceived = (double)this.nrinterestsreceived;
			this.popularity =(datasent/intreceived);
		}
	}
	
	/**
	 * Updates delivery predictions for a host.
	 * <CODE>P(a,content_b) = P(a,content_b)_old + (1 - P(a,content_b)_old) * P_INIT</CODE>
	 * @param host The host we just met
	 */
	private void updateDeliveryPredFor(DTNHost to) {
		
		Double oldValue = 0.0;
		MessageRouter otherRouter = to.getRouter();
		Map<String, Integer> otherContents = ((PIFP_Proactive)otherRouter).lc;
		
		for (String name : otherContents.keySet())
		{
			//check if the content already exists in the CS itself
			
			if (!(this.bloomLC.contains(name)) && !(checkifExistesDeliveredmessages(name)))
			{
				oldValue = getPredFor(name);
				if (oldValue == null) oldValue = 0.0;
				double newValue = oldValue + (1 - oldValue) * P_INIT; 
				this.FIB.put(name, newValue);
				
				if(this.bloomPIT.contains(name)) 
				{
					for (Message m : this.getMessageCollection()) 
					{
						 String datas [] = m.getData().split(" ");
						 
						if (to.getRouter().hasMessage(m.getId())) 
						{
							continue; // skip messages that the other one has
						}
						
						if (datas[0].equalsIgnoreCase("pi"))
						{
							if (datas[1].equalsIgnoreCase(name))
							{
								if(sendMessage(m.getId(), to) <1)
								{ 
									if(I.LI.containsKey(m.getId()))
									{
										if(!(I.getisSendOutInterest(m.getId())))
										{
										 I.setTimeSendOutInterest(m.getId());
										 I.setisSendOutInterest(m.getId(), true);								 
										 I.incrSendOutInterests();
										} 
									}
								}		 
							 }
						 }
					}
				}	
			}
		}
		return;
	}

	/**
	 * Updates transitive (A->B->Content_C) delivery predictions.
	 * <CODE>P(a,content_c) = P(a,content_c)_old + (1 - P(a,content_c)_old) * P(a,b) * P(b,content_c) * BETA
	 * </CODE>
	 * @param host The B host who we just met
	 */
	private void updateTransitivePreds(DTNHost host) {
		
		MessageRouter otherRouter = host.getRouter();
		double pa_c_old =0;
		double  pa_b = 0.75;	 		  
										 
		// get predictabilities the other host
		Map<String, Double> otherPreds = ((PIFP_Proactive)otherRouter).getDeliveryPreds();
		
		Map<String, Integer> otherContent = ((PIFP_Proactive)otherRouter).lc;
		
		for (String e : otherPreds.keySet()) {
			// verifies that does not contain any existing content in CS
			if (this.bloomLC.contains(e) || checkifExistesDeliveredmessages(e) || otherContent.containsKey(e)) // verifica se no contm nenhum contedo do qual ele possui
			{
				continue;
			}

			double pb_c = 0;
			pb_c = otherPreds.get(e); // get the predictability of node b reach the content c
			pa_c_old = getPredFor(e); // P(a,c)_old - predictability of node a reach the content c
			double pNew = pa_c_old + ( 1 - pa_c_old) * pa_b * pb_c * beta; //p(a,c) = P(a,c)old + (1-P(a,c)old).P(a,b).P(b,c)xBeta
		
			this.FIB.put(e, pNew);
			
			if(!this.bloomFIB.contains(e)) 
			{
				this.bloomFIB.add(e);
			}

		}
	}

	
	
	/**
	 * Returns the current prediction (P) value for a host or 0 if entry for
	 * the host doesn't exist.
	 * @param host The host to look the P for
	 * @return the current P value
	 */
	
	public double getPredFor(String name) {
		ageDeliveryPreds(); // make sure preds are updated before getting
		if (this.FIB.containsKey(name)) {
			return this.FIB.get(name);
		}else
		{
			return 0;
		}
		
	}
	


	/**
	 * Ages all entries in the delivery predictions.
	 * <CODE>P(a,b) = P(a,b)_old * (GAMMA ^ k)</CODE>, where k is number of
	 * time units that have elapsed since the last time the metric was aged.
	 * @see #SECONDS_IN_UNIT_S
	 */
	private void ageDeliveryPreds() {
		Double value; 
		double timeDiff = (SimClock.getTime() - this.lastAgeUpdate) / secondsInTimeUnit;

		
		
		if (timeDiff == 0){
			this.lastAgeUpdate = SimClock.getTime();
			return;
		}
		
		double mult = Math.pow(GAMMA, timeDiff);
		
		if (this.FIB.isEmpty()){
				return;
		}
		
		for (String e : this.FIB.keySet()){
			value = this.FIB.get(e);
			this.FIB.put(e, mult*value);  
		}
		
		this.lastAgeUpdate = SimClock.getTime();
	}
	
	/**
	 * Returns a map of this router's delivery predictions
	 * @return a map of this router's delivery predictions
	 */
	private HashMap<String, Double> getDeliveryPreds() {
		ageDeliveryPreds(); // make sure the aging is done
		agepopularity();
		return this.FIB;
	}

	
	/**
	 * Remove set of messages
	 * @param ArrayList Messages
	 */
	public void removemessages(ArrayList<String> messages)
	{			
		for (String Id : messages)
		{
			this.deleteMessage(Id, true);
		}
	}
		
	
	/**
	 * Called when it receives data and possesses pending interest messages.
	 * A message of interest is eliminated, created by the node itself.
	 * @param name is the name of content;
	 * @return "removed", if a message of interest is removed
	 */	
	public String removeInterest(String name, boolean setSatisfyPendingInterest)
	{
		Collection<Message> msgCollection = getMessageCollection();	
		
		for (Message m : msgCollection) 
		{
			String id;
			String data = m.getData();
			String datas[] = data.split(" ");
			
			if(datas[1].equalsIgnoreCase(name))
			{
				if(datas[0].equalsIgnoreCase(INTEREST_PACKAGE))
				{
					id = m.getId();

					if((m.getFrom() == this.getHost()))
					{
						this.deleteMessage(m.getId(), true);
						return null;
					}
					
				}
			}
		}
		
	

		return null;	
	}
	
	
	
	
	// com CS oportunista 
	
	/*public void onData(Message m, String name, DTNHost from){	
		
		ArrayList<DTNHost> allHosts = new ArrayList<DTNHost>();
		
		if (bloomPIT.contains(name))
		{
			if (this.PIT.containsKey(name))
			{
				allHosts = this.PIT.get(name).getAllHosts();
				
				// esta parte  para encaminhar os dados recebidos pelo mesmo caminho que recebeu interesse
				
				for (DTNHost host: allHosts)
				{
					if(this.getHost() == host) continue;
					else 
					{
						if(this.lei.containsKey(name))
						 {
							 if(!this.lei.get(name).contains(host))
								 this.lei.get(name).add(host);
						 }
						 else {
							 ArrayList<DTNHost> aux = new ArrayList<DTNHost>();
							 aux.add(host);
							 this.lei.put(name, aux);
						 }
						
					}
				}
				

				if(allHosts.contains(this.getHost())) // se o n for subscritor do conteudo tbm
				{	
					if (lc.containsKey(name))
					{
							if (this.lc.get(name)!=1) this.lc.put(name, 2); // adiciona a CS o conteudo
					}else  this.lc.put(name, 2); // adiciona a CS o conteudo
					
														 
					System.out.println("No:" + this.getHost() + "recebeu dados: " + name + "de: " + m.getFrom()); 
					
					System.out.println("Armazena mensagem ");
					
					//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
					
					// Significa que agora j tem na LC o contedo 
					// para os outros ns no calcularem a transitividade mas sim a presivilidade -.-
					if (this.bloomFIB.contains(name))
					{
						this.bloomFIB.remove(name);
						this.FIB.remove(name);
					}
					
					deliveredMessages.put(m.getId(), m);
					
					//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
					
					
					//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
					// 	Se no tiver outras interfaces para enviar (faces)
					if(allHosts.size() <= 1)
						removeFromMessages(m.getId());
					else m.setFrom(this.getHost());
					
					// remove da PIT o nome do interesse
					String idm[] = m.getId().split("_");
					bloomPIT.remove(name);
					this.PIT.remove(name);	
					removeInterest(name, true);
					
					
					// RELATRIOS
					// define o tempo de chegada de dados ou pode-se dizer que  quando o interesse  satisfeito
					if (I.LI.containsKey(idm[1])){
						I.settimeSastifiedInterest(idm[1]);
						I.setisSatisfiedInterest(idm[1], true);
					}
					
					// criar lista local que serve para guardar o instante de quando  enviado o primeiro e  satisfeito
					
					R.timeDelivered(m.getId());
					R.setDeliveredData(m.getId(),  true);
					R.nrofhops = m.getHopCount();
				}
				else 
				{			
					m.setFrom(this.getHost());
					System.out.println("Armazena Dados em Buffer com id: " + m.getId());
					if(!this.lc.containsKey(name)) this.lc.put(name, 3); // adiciona a CS o conteudo
					this.bloomLC.add(name);
				
					// remove interesse pois ja recebeu dados 
					bloomPIT.remove(name);
					
					if (bloomFIB.contains(name))
					{
						bloomFIB.remove(name);
						this.FIB.remove(name);
					}
						
					this.PIT.remove(name);
					removeInterest(name, false);
					//agora devia preparar-se para enviar os dados para as interfaces que pediram 
				}
			}
			else {
				this.PIT.remove(name);
				bloomPIT.remove(name);
				removeInterest(name, false);
				m.setTtl(20);
				
			}
		}
		m.setTtl(20); //  como se fosse eliminado

	}*/
	

	/**
	 * Processing the data packets received from the network. 
	 * If the node is subscriber data is added to the repository. 
	 * If not are just kept in the CS temporarily.
	 * @param m is message
	 * @param name is the name of content
	 * @param from is the sender
	 */
	public void onData(Message m, String name, DTNHost from){	
		
		ArrayList<DTNHost> allHosts = new ArrayList<DTNHost>();
		boolean isubscriber = false;
		
		if (this.PIT.containsKey(name))
		{
				allHosts = this.PIT.get(name).getAllHosts(); // get all nodes interested in the data

				if(allHosts.contains(this.getHost()))
				{
					String idm[] = m.getId().split("_");	

					createCSEntrie(name, m, 2); // create Content Store entry
					I.createPITSast(this.PIT.get(name).timeCreated, SimClock.getTime());
					deliveredMessages.put(m.getId(), m); // add message to repository 

					deletePITEntrie(name);
						
					R.nrDeliveredData(); 
					I.incrSastifiedInterests();
					R.incrNrHops(m);

					removeInterest(name, true); 
							
					// Reports ------------------
					if (I.LI.containsKey(idm[1])){
						I.settimeSastifiedInterest(idm[1]);
						I.setisSatisfiedInterest(idm[1], true);
					}
							
					if (R.PC.containsKey(m.getId()))
					{	
						R.timeDelivered(m.getId());
						R.setDeliveredData(m.getId(),  true);
					}
						
				}
				else 
				{
					if(this.lei.containsKey(name))
					{
						if(!this.lei.get(name).contains(host))
						this.lei.get(name).add(host);
					}
					else 
					{
						ArrayList<DTNHost> aux = new ArrayList<DTNHost>();
						aux.add(host);
						this.lei.put(name, aux);
					}
				}
				if(isubscriber!=true)
				{
					//Add CS entry
					if(!this.lc.containsKey(name)) 
					{
						createCSEntrie(name, m, 3); 
					}
				
					deletePITEntrie(name);
					removeInterest(name, false);
				}
			
		}
		else 
		{
			if(!this.lc.containsKey(name)) this.lc.put(name, 3); 
			if(!this.bloomLC.contains(name)) this.bloomLC.add(name);
		}
	}
	
	
	
	
	public void sendDataMessage(Message c, String name, NDNRouter othRouter)
	{			
		 if(this.lei.containsKey(name))
		 {
			 if(!this.lei.get(name).contains(othRouter.getHost()))
				 this.lei.get(name).add(othRouter.getHost());
		 }
		 else {
			 ArrayList<DTNHost> aux = new ArrayList<DTNHost>();
			 aux.add(othRouter.getHost());
			 this.lei.put(name, aux);
		 }
		
	}
	
	
	
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	public void addInterest(String name, int nrOfRequests, boolean isSubs, String content){
		InterestList newinterest = new InterestList();
		
		if(!li.containsKey(name)){
			newinterest = new InterestList(name,nrOfRequests,isSubs,0, SimClock.getTime());
			li.put(name, newinterest);
		}
	
	}
	
	/**
	 * Updates all popularities in accordance with the time spent
	 * <CODE>Popularity_interestx = Popularity_interestx_old * (GAMMA ^ timeDiff)</CODE>, where timeDiff is number of
	 * time units that have elapsed since the last time the metric was aged.
	 * @see #SECONDS_IN_UNIT_S
	 */
	public void agepopularity()
	{
		double timeDiff = (SimClock.getTime() - this.lastupdate) / secondsInTimeUnit;

		if (timeDiff == 0){
			return;
		}
		double mult = Math.pow(GAMMA, timeDiff);
		
		for(InterestList i: li.values()){
			i.setPopularity(i.getPopularity()*mult); //update popularity
			li.put(i.getName(), i);
		}
		
		this.lastupdate = SimClock.getTime();
	}
	
	
	
	/**
	 * Sort by popularity to send the contents
	 * @param messages to send 
	 * @return the list of messages sorted in decreasing order
	 */
	private List<Tuple<Message, Connection>> orderPopularity(List<Tuple<Message, Connection>> messages){
		TreeMap<String,Tuple<Message, Connection>> list = new TreeMap<String,Tuple<Message, Connection>>();
		String data;
		String[] datas;
		InterestList newInterest = new InterestList();
		int i=0;
		
		for(Tuple<Message, Connection> t : messages){
			data = t.getKey().getData();
			datas = data.split(" ");
			if(this.li.containsKey(datas[1]))
			{
				newInterest = li.get(datas[1]);	
				String popularity = String.valueOf(newInterest.getPopularity())+"pop"+i;
				list.put(popularity, t);
				i++;			
			}	
		}
		List<Tuple<Message, Connection>> list2 = new ArrayList<Tuple<Message, Connection>>();
		list2.addAll(list.descendingMap().values());

		return list2;
	}
	

	
	
	public void updatePopularity(){
		int nrOfRequests=0;
		double popularity;
		Map<String, InterestList> temp = new HashMap<String, InterestList>();
		temp.putAll(li);
		
		
		if(!li.isEmpty()){
			for(InterestList i: temp.values()){
				nrOfRequests += i.getNrOfRequests(); 
			}
			for(InterestList i: temp.values()){
				if(nrOfRequests==0){
					popularity=0;
				}
				else
					popularity = Double.valueOf(i.getNrOfRequests()) / Double.valueOf(nrOfRequests);
				i.setPopularity(popularity);
				li.put(i.getName(), i);
			}
		}
	}
	
	
	
	/**
	 * Reception of a interst packet that is treated. this time the packet is analyzed
	 * and the conditions for acceptance of new interest are checked. 
	 * If any content that satisfies the interest is returned immediately, 
	 * otherwise the PIT is added a new entry with the conditions.
	 * @param m is message
	 * @param name is the name of content
	 * @param from is the sender
	 */
	
	public void onInterest(Message m, String name, DTNHost from)
	{	
		String content = "";
		String[] data = name.split("/");
		int size = data.length;
		content = data[size-1];
		NDNRouter othRouter = (NDNRouter)from.getRouter();
		Message c = null;
		
		
		if(!this.li.containsKey(name)) // adds interest to a list of all the interests that the node has knowledge
		{
			addInterest(name,1,false, null);
			updatePopularity();
		}
		else{
			InterestList newInterest = new InterestList();
			newInterest = this.li.get(name);
			newInterest.setNrofRequests((newInterest.getNrOfRequests()) + 1);
			this.li.put(name, newInterest);
			updatePopularity();
		}		
		
		if (this.bloomLC.contains(name)) // se existir conteudo enviar 
		{
			if(I.existsIdMessage(m.getId()) == false) I.incrNrHops(m);

			this.nrinterestsreceived = this.nrinterestsreceived + 1;
			
			
			if(!this.hasMessage(RESPONSE_PREFIX+m.getId()) && !this.deliveredMessages.containsKey(RESPONSE_PREFIX+m.getId()))
			{
				c = new Message(this.getHost(), othRouter.getHost(), RESPONSE_PREFIX+m.getId(), drawMessageSize(), CONTENT_PACKAGE+ " " + name + " " +name);

				if(createNewMessage(c))
				{
					//send PC
					c.setTtl(500);
					sendMessage(m.getId(),from);
					sendDataMessage(c, name,  othRouter);
				}		 
			}
			else 
			{
				if(!this.hasMessage(RESPONSE_PREFIX+m.getId()))
				{
					if (this.deliveredMessages.containsKey(RESPONSE_PREFIX+m.getId()))
					{
						Message msg = this.deliveredMessages.get(RESPONSE_PREFIX+m.getId());
						msg.setTo(othRouter.getHost());
						addToMessages(msg, false);
						sendMessage(m.getId(),from);
						sendDataMessage(msg, name, othRouter);
					}
				}
				else 
				{
					Collection<Message> msgCollection = getMessageCollection();
					for(Message d : msgCollection )
					{
						if (d.getId().equalsIgnoreCase(RESPONSE_PREFIX+m.getId()))
						{
							 d.setTo(othRouter.getHost());
							 sendMessage(m.getId(),from);
							 sendDataMessage(d,  name, othRouter);
							 break;
						}
					}	
				}
			}
			if(I.LI.containsKey(m.getId())) 
			{	
				I.settimeDelivered(m.getId());
			}
		}
		else if(bloomFIB.contains(name))
		{
				this.nrinterestsreceived = this.nrinterestsreceived + 1;
				
				insertPI(m, false, from);
				
				if (!bloomPIT.contains(name)) 
				{
					bloomPIT.add(name);
				}
		
		}
		else
		{  //If bloom FIB entry has no interest to rule out interest
			this.toremove.put(m.getId(), SimClock.getTime());
		}
		
	}
	
	
	
	/**
	 * Try send Data to other router
	 * @return Tuple<Message, Connection>
	 */
	
	private Tuple<Message, Connection> trySendData() {

		List<Tuple<Message, Connection>> outgoingMessages = new ArrayList<Tuple<Message, Connection>>(); 
		
		Collection<Message> msgCollection = getMessageCollection();
			
		/* for all connected hosts collect all messages that have a higher
			  probability of delivery by the other host */
		for (Connection con : getConnections()) 
		{
			outgoingMessages.clear();
			DTNHost other = con.getOtherNode(getHost());
			NDNRouter othRouter = (NDNRouter)other.getRouter();
				
				
			if (othRouter.isTransferring()) {
				continue; // skip hosts that are transferring
			}
				
			for (Message o : msgCollection)
			{
				String data = o.getData();
				String[] datas = data.split(" ");
					
				if(datas[0].equalsIgnoreCase(CONTENT_PACKAGE)) {
						
					if (othRouter.hasMessage(o.getId())) {
						continue; // skip messages that the other one has
					}
						
					if (othRouter.bloomLC.contains(datas[1])) {
						continue; // skip messages that the other one has
					}
						
					if (!outgoingMessages.contains(o))
					{
						outgoingMessages.add(new Tuple<Message, Connection>(o,con));
					}
				}
				
		
			}
			
			if (outgoingMessages.size() == 0) {
				return null;
			}
			
			outgoingMessages = orderPopularity(outgoingMessages);

			Tuple<Message, Connection> tupleSend = tryMessagesForConnected(outgoingMessages);	// try to send messages
				
		
			if(tupleSend!=null)
			{		
				Message send = tupleSend.getKey(); 
				Connection c = tupleSend.getValue();
				DTNHost h = c.getOtherNode(getHost());
				String[] dat = send.getData().split(" ");
						
				if (!(R.PC.containsKey(send.getId()))) R.createNewPC(send.getId(), (send.getData().split(" "))[1], true);
					
				R.incrSendOutData();
				this.nrdatasent = this.nrdatasent +1;
							
				if(this.lei.containsKey(dat[1]))
				{
					if(this.lei.get(dat[1]).size()<=1) // se for o unico para remover elimina a entrada por completo 
					{
						this.lei.remove(dat[1]);
					}
					else  this.lei.get(dat[1]).remove(h); // se no for o unico remove apenas o n para quem enviar 
				}		
			}
		}
		return null;
	}

	
	

	/**
	 * Try send Data to other router
	 * @return Tuple<Message, Connection>
	 */
	public void checkmensagensSatisfeitas()
	{
		Collection<Message> msgCollection = this.getMessageCollection();
		ArrayList<String> idtoRemove = new ArrayList<String>() ;

	
		for (Message m : msgCollection) 
		{
			String data = m.getData();
			String[] datas = data.split(" ");
		
			
			if (datas[0].equalsIgnoreCase(INTEREST_PACKAGE)) 
			{
				
				if(this.lc.containsKey(datas[1]))
				{
					deletePITEntrie(datas[1]);

					if(m.getFrom() == this.getHost())
					{
						I.numberInterestsSastifiedinNode();
							
						if(!idtoRemove.contains(m.getId()))
						{
							idtoRemove.add(m.getId());	
						}
					}else
					{
						if(!idtoRemove.contains(m.getId()))
						{
							idtoRemove.add(m.getId());	
								
						}
					}
				}	
			}
		}
	removemessage(idtoRemove);
	}
	
	
	/**
	 * Try send Interest to other router
	 * @return Tuple<Message, Connection>
	 */
	private Tuple<Message, Connection> trySendInterest() 
	{
	ArrayList<String> idtoRemove = new ArrayList<String>() ;
					
	List<Tuple<Message, Connection>> outgoingMessages = new ArrayList<Tuple<Message, Connection>>(); 
			
	List<Tuple<Message, Connection>> out = new ArrayList<Tuple<Message, Connection>>(); 
		
	Collection<Message> msgCollection = getMessageCollection();
	/* for all connected hosts collect all messages that have a higher
	 probability of delivery by the other host */
			
	checkmensagensSatisfeitas();
			
	for (Connection con : getConnections()) 
	{
		idtoRemove.clear();
		outgoingMessages.clear();
		out.clear();
				
		DTNHost other = con.getOtherNode(getHost());
		NDNRouter othRouter = (NDNRouter)other.getRouter();
				
				
		for (Message m : msgCollection) 
		{
			String data = m.getData();
			String[] datas = data.split(" ");
					
			if (othRouter.hasMessage(m.getId())) 
			{	
				continue; // skip messages that the other one has
			}

			if (datas[0].equalsIgnoreCase(INTEREST_PACKAGE)) 
			{
				if(this.bloomPIT.contains(datas[1]))
				{
					if(othRouter.bloomLC.contains(datas[1])) // se o outro host possuir na sua lc enviar
					{
						out.clear();
						out.add(new Tuple<Message, Connection>(m,con));				
						Tuple<Message, Connection> res = tryMessagesForConnected(out);

						if(res!=null)
						{			
							if(!I.getisSendOutInterest(res.getKey().getId()))
							{
								I.setisSendOutInterest(res.getKey().getId(), true);
								I.setTimeSendOutInterest(res.getKey().getId());
								I.incrSendOutInterests();
							}
						}
					}	
					else  if (othRouter.bloomFIB.contains(datas[1])) 
					{
						if (comparePrevs(this, othRouter, datas[1])>=1)
						{	
							outgoingMessages.add(new Tuple<Message, Connection>(m,con));
						}
					}		
				}
			}
		}
			
		
				
		if (outgoingMessages.size() == 0) {
			return null;
		}		
					
		//sort messages in descending order according to the predictability
		outgoingMessages = orderByPrev(outgoingMessages);
		//try to send messages to the corresponding connections
		Tuple<Message, Connection> tupleSend = tryMessagesForConnected(outgoingMessages);
					
		if(tupleSend!=null)
		{
			
			Message send = tupleSend.getKey(); 
			//Reports
			if(!I.getisSendOutInterest(send.getId()))
			{
				I.setisSendOutInterest(send.getId(), true);
				I.setTimeSendOutInterest(send.getId());
				I.incrSendOutInterests();
			}
		}
	}
	return null; 	
	}

		
	/**
	 * Remove set of messages
	 * @param ArrayList Messages
	 */
	private void removemessage(ArrayList<String> messages)
	{			
		for (String Id : messages)
		{
			this.deleteMessage(Id, true);
		}
	}
		
	/**
	 * Order Set of Messages By Probability of reach the producer
	 * @param List<Tuple<Message, Connection>>  Messages
	 * @return  List<Tuple<Message, Connection>> Messages ordered by Probability
	 */
		
	private List<Tuple<Message, Connection>> orderByPrev(List<Tuple<Message, Connection>> messages)
	{
			
		TreeMap<String,Tuple<Message, Connection>> list = new TreeMap<String,Tuple<Message, Connection>>();	
		String data;
		String[] datas;
		int i=0;
			
		for(Tuple<Message, Connection> t : messages)
		{
			data = t.getKey().getData();
			datas = data.split(" ");
			double prev;
				
			if (this.FIB.containsKey(datas[1]))
			{
				prev = this.FIB.get(datas[1]);
			}
			else prev=1;
				
			String pred = String.valueOf(prev+"pop"+i);
			list.put(pred, t);
			i++;
		}
			
		List<Tuple<Message, Connection>> list2 = new ArrayList<Tuple<Message, Connection>>();
		list2.addAll(list.descendingMap().values());

		return list2;	
	}
			
	
	
	@Override
	public RoutingInfo getRoutingInfo() {
		ageDeliveryPreds();
		agepopularity();
		RoutingInfo top = super.getRoutingInfo();
		RoutingInfo ri = new RoutingInfo(FIB.size() + 
				" delivery prediction(s)");
		
		for (Entry<String, Double> e : FIB.entrySet()) {
			String name = e.getKey();
			Double value = e.getValue();
			
			ri.addMoreInfo(new RoutingInfo(String.format("%s : %.6f", 
					name, value)));
		}
		
		top.addMoreInfo(ri);
		return top;
	}
	
	
	public int comparePrevs(NDNRouter r1, NDNRouter r2, String name)
	{
		double p1; // este router
		double p2; // outro router
		
		double lastProb= this.PIT.get(name).getLastProb();
		
		if (r1.FIB.containsKey(name)) p1 = r1.FIB.get(name);
		else p1 = 0;
		
		if (r2.FIB.containsKey(name)) p2 = r2.FIB.get(name);
		else p2=0;
		
		// bigger probability should come first
		if (p2-p1 == 0) {
			/* equal probabilities -> let queue mode decide */
			//return compareByQueueMode(tuple1.getKey(), tuple2.getKey());
			
			if (p2-lastProb > 0)
			{
				this.PIT.get(name).setLastProb(p2);
				return 1;
			}
			else return -1;
		}
		else if (p2-p1 < 0) {
			return -1;
		}
		else {
			if (p2-lastProb > 0)
			{
				this.PIT.get(name).setLastProb(p2);
				return 1;
			}
			else return -1;
		}		
		
	}

	
	

	@Override
	public PIFP_Proactive replicate() {
		return new PIFP_Proactive(this);
	}
	
	
	@Override
	public void update() {
		super.update();
		if (isTransferring() || !canStartTransfer()) {
			return; // can't start a new transfer
		}
	
		this.trySendInterest(); // try send interest first
		this.trySendData();	// try send data 
		
	}

}


