/* 
 /* 
 * Copyright 2015 Universidade do Minho, GCOM
 * Modified by Paulo Duarte.
 * Released under GPLv3. See LICENSE.txt for details. 
 */

package routing;


import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import ca.pfv.spmf.patterns.cluster.Cluster;
import ca.pfv.spmf.patterns.cluster.DoubleArray;
import routing.NDN.digest;
import routing.NDN.socialTie;
import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;
import core.SimClock;
import core.Tuple;

public class sctr extends sctrRouter{

	protected int[] hostRange = {0, 220};
	
	protected Map<String, Double> last_relay_node; 
	
public sctr(Settings s) {
		
		super(s);
		
		// TODO Auto-generated constructor stub
	}
	
	public sctr(sctr r)
	{
		
		super(r);
		last_relay_node = new HashMap<String, Double>();

	}


	@Override
	public void changedConnection(Connection con) {
		
		if (con.isUp()) {
			
			// trocar tabela de enaminhamento  
			DTNHost otherHost = con.getOtherNode(getHost());
			
			Message m = null;

			recordTimestamp(otherHost);
			
			computeSocialRelationship(otherHost);
			updateSocialTieTable();
			
			convergeContentTable(otherHost);
			convertDigestTable(otherHost);
			
			//s vai atualizar se o outro n tiver uma centralidade maior 
			exhangeSocialTie(otherHost);
			
			ComputeCentrality();
			mycentralitie();
			
			if(this.getHost().getConnections().size() <= 1)
			{
				checkcentralitie();
			}
			
			// 1 guarda o instante de tempo em que foi estabelecido contato com o n 
			
		}
		
		
	}
	
	

	@Override
	public Message messageTransferred(String id, DTNHost from) {
		Message m = super.messageTransferred(id, from); 
		String data = m.getData();
		
		/** 
		 * datas[0] = type
		 * datas[1] = name
		 * datas[2] = content
		 */
		
		String[] datas = data.split(" ");
		
		System.out.println( "" + this.host + " recebeu-> " + datas[0]+ " -> " + datas[1] + " de: " + m.getFrom());
		
		
		//Caso receba PI
		if(datas[0].equalsIgnoreCase(INTEREST_PACKAGE)){			
			onInterest(m);
			// aqui posso atualizar os clusters
		}
		
		//Caso receba PC
		if(datas[0].equalsIgnoreCase(CONTENT_PACKAGE)){
			onData(m);
			
		}
		
	
		return m;
	}
	
	// dado um n  calculara a relao social com esse n
	// Ri(J)= somatrio F(tbase - Tjk)
	// procurar no encounter-vector  o n
	// percorrer e ver o instante 
	// Calcular F(x)


	public void computeSocialRelationship(DTNHost h)
	{
		double RiJ = 0.0;
		double x;
		double tBase = SimClock.getTime(); 
		
		if (this.encounterVector.containsKey(h))
		{
			ArrayList<Double> times = this.encounterVector.get(h);
			
			for (Double t : times)
			{
				x =  (tBase - t);
				RiJ = RiJ + Fx(x);
			}
		}
		
		socialTie socialT= new socialTie();
		socialT.idNode = h;
		socialT.timestamp = tBase; 
		socialT.socialvalue= RiJ;
		socialT.thisidNode = this.getHost();
		
		socialTieTable.put(h, socialT);
		
	}
	

	// atualiza a relao do proprio com os outros ns 
	
	public void updateSocialTieTable()
	{
		double RiJ = 0.0;
		int x;
		double tBase = SimClock.getTime(); 
		double fx;
		
		for ( DTNHost h: encounterVector.keySet())
		{
			
				ArrayList<Double> times = this.encounterVector.get(h);
				
				for (Double t : times)
				{
					x = (int) (tBase - t);
					fx = Fx(x);
					RiJ = RiJ + fx;
				}
			
			if(socialTieTable.containsKey(h))
			{
				socialTieTable.get(h).socialvalue = RiJ;
				socialTieTable.get(h).timestamp = tBase;
			}
			else{
				socialTie socialT= new socialTie();
				socialT.idNode = h;
				socialT.thisidNode = this.getHost();
				socialT.timestamp = tBase; 
				socialT.socialvalue= RiJ;
				socialTieTable.put(h, socialT);
			}
			
		}
			
		
	}
	
	public double Fx(double x)
	{
		// 1/2 x ^(x.1exp^-4)
		double lamb = 0.001;
		return Math.pow(0.5, (lamb*x));
	}
	
	
	public void recordTimestamp(DTNHost h)
	{
		ArrayList<Double> times = new ArrayList<Double> ();
		
		if (encounterVector.containsKey(h))
		{
			encounterVector.get(h).add(SimClock.getTime());
		}
		else
		{
			times.add(SimClock.getTime());
			encounterVector.put(h, times);
		}

	}
	
	
	public void exhangeSocialTie(DTNHost other)
	{
				
		MessageRouter otherRouter = other.getRouter();
		
		
		// obtem as tabelas de relaes social 
		
		Map<String, socialTie> otherSocialTie = 
				((sctrRouter)otherRouter).socialTieTableOthers;  // tabela dos outros para outros 
		
		Map<DTNHost, socialTie> otherSTie = ((sctrRouter)otherRouter).socialTieTable; // tabela dele para os outros

		// percorre a tabela dos outros para os outros 
		for (String key : otherSocialTie.keySet())
		{
			if(otherSocialTie.get(key).thisidNode!= this.host)
			{
				if(this.socialTieTableOthers.containsKey(key))
				{
					if(otherSocialTie.get(key).timestamp > this.socialTieTableOthers.get(key).timestamp) // se timestamp do outro for mais recente atualizar
					{
						this.socialTieTableOthers.get(key).socialvalue = otherSocialTie.get(key).socialvalue;
						this.socialTieTableOthers.get(key).timestamp = otherSocialTie.get(key).timestamp;
					}
				}
				else 
				{
					this.socialTieTableOthers.put(key, otherSocialTie.get(key));
				}
			}

		}
		
		
		
		for(DTNHost host : otherSTie.keySet() )
		{	
			String auxKey = other.toString() + "_" + host.toString();
			this.socialTieTableOthers.put(auxKey,otherSTie.get(host));
		}
		
		
		
	}
	
	
	//ci 
	
	public double ComputeCentrality()
	{
		
		double c =0.0;
		int N = 0;
		double R=0.0;
		DTNHost host = null; 
		
		for (int h : hostRange)
		{
			for (String key : socialTieTableOthers.keySet())
			{
				if( socialTieTableOthers.get(key).idNode.getAddress() == h)
				{
					N = N +1;
					R = R +  socialTieTableOthers.get(key).socialvalue;
					host = socialTieTableOthers.get(key).thisidNode;
				}
				
			}
			
			if(host!=null)
			{
				c = (alpha*R)/N + ((1-alpha))*(R*R)/(N*(R*R));
				centralitie.put(host, c);
				host = null;
			}
			
		}
		return R;
		
	}
	
	
	public void mycentralitie()
	{
		int N = 0;
		double R = 0;
		double c;
		
		for (DTNHost key : this.socialTieTable.keySet())
		{
				N = N +1;
				R = R + socialTieTable.get(key).socialvalue;
		}
		if(R!=0)
		{
			c = (alpha*R)/N + ((1-alpha))*(R*R)/(N*(R*R));
			centralitie.put(this.host, c);
		}
		
	}

	
	// quando ele decidir enviar um interesse ou trocar a tabela de contents 
	// calcula a
	public void computeCluster(int address)
	{
		
	}
	
	
	
	public void onInterest(Message m)
	{
		List<Tuple<Message, Connection>> outgoingMessages = 
				new ArrayList<Tuple<Message, Connection>>(); 
		

		
		String data = m.getData();
		String [] datas = data.split(" ");
		
		System.out.println();
		
		if(m.getTo() == this.host) // quer dizer que foi direcionada para o provedor 
		{
			if (this.lc.containsKey(datas[1])) System.out.println("tem dados ... ");
			if(this.bloomLC.contains(datas[1])) // verifica se contm na LC O CONTEDO 
			{
				Message c = new Message(this.getHost(), m.getFrom(), RESPONSE_PREFIX+m.getId(), drawMessageSize(), CONTENT_PACKAGE+ " " + datas[1] + " " +datas[1]);
				
				if(createNewMessage(c)){
					// os PC seguem o mesmo algoritmo que os pacotes de interesse
					//sendDataMessage(c, name,  othRouter);
					//removeFromMessages(m.getId());
					trysendData();
				}
			}
			
			// arranjar maneira de dizer que o conteudo j foi enviado 
		}
		else 
		{
			if(m.getTo()==null)
			{
				for( DTNHost h: this.digestTable.keySet())
				{
					for (String name: this.digestTable.get(h).dig.keySet())
					{
						if (name.equalsIgnoreCase(datas[1]))
						{
							m.setTo(h);
							break;
						}
					}
				}
			}
		}
		
		// ver a lista de contedos
		// se existir 
		// ver qual o provedor de contedo
		// senao continua a enviar o pacote de interesse de acordo com a estratgia 
		
	}
	
	public void onData(Message m) // enviado dependendo do social-tie level 
	{
		
		List<Tuple<Message, Connection>> outgoingMessages = 
				new ArrayList<Tuple<Message, Connection>>(); 
		
		ArrayList<String> idtoRemove = new ArrayList<String>() ;
		
		
		String data = m.getData();
		String [] datas = data.split(" ");
		
		if(m.getTo() == this.host) // quer dizer que chegou uma mensagem de conteudo ao consumidor
		{
			
			if (this.PIT.containsKey(datas[1]))
			{
				
				// remove da PIT o nome do interesse
				String idm[] = m.getId().split("_");
			
				removeFromMessages(m.getId());
				
				System.out.println("Recebeu DADOS!! ");
				// RELATRIOS
				// define o tempo de chegada de dados ou pode-se dizer que  quando o interesse  satisfeito
				if (I.LI.containsKey(idm[1])){
					I.settimeSastifiedInterest(idm[1]);
					I.setisSatisfiedInterest(idm[1], true);
				}
				
				// criar lista local que serve para guardar o instante de quando  enviado o primeiro e  satisfeito
				this.PIT.remove(datas[1]);
				this.bloomLC.add(datas[1]);
				this.lc.put(datas[1], datas[1]);
				idtoRemove.add(m.getId());
				
				I.incrSastifiedInterests();
				R.nrDeliveredData(); 
				
				R.timeDelivered(m.getId());
				R.setDeliveredData(m.getId(),  true);
			//	R.nrofhops = m.getHopCount();
			}
			// arranjar maneira de dizer que o conteudo j foi enviado 
		}
		else 
		{
			trysendData();
		}
		
		
	}
	
	
	public void trysendData()
	{
		
		List<Tuple<Message, Connection>> outgoingMessages = 
			new ArrayList<Tuple<Message, Connection>>(); 

		Collection<Message> msgCollection = getMessageCollection();
		/* for all connected hosts collect all messages that have a higher
		   probability of delivery by the other host */
		
		List<Cluster> clusters = null;
		
		for (Connection con : getConnections()) {
			
			DTNHost other = con.getOtherNode(getHost());
			sctrRouter othRouter = (sctrRouter)other.getRouter();
			
			
			if (othRouter.isTransferring()) {
				continue; // skip hosts that are transferring
			}
			
					
			for (Message m : msgCollection) {
				
				
				if (othRouter.hasMessage(m.getId())) {	
					continue; // skip messages that the other one has
				}
				
				
				String split[] = m.getData().split(" ");
				String interestName = split[1];
				
				
				String data = m.getData();
				String[] datas = data.split(" ");
				String name = datas[1];
				
				if (datas[0].equalsIgnoreCase(CONTENT_PACKAGE)) {
					// ver se o n existe na tabela de encontros
					if(m.getTo() == other)
					{	
						outgoingMessages.add(new Tuple<Message, Connection>(m,con));
					}
					else
					{
						for (String key : this.socialTieTableOthers.keySet())
						{
							if (this.socialTieTableOthers.get(key).thisidNode == other) // origem 
							{
								if(this.socialTieTableOthers.get(key).idNode == m.getTo()) // para o destino
								{
									if (this.socialTieTable.containsKey(m.getTo()))
									{
										if (this.socialTieTableOthers.get(key).socialvalue > this.socialTieTable.get(m.getTo()).socialvalue)
										{
											outgoingMessages.add(new Tuple<Message, Connection>(m,con));
											// enviar 
										}
									}
									
								}
							}	
						}
					
					}

				}
			}
			if(!outgoingMessages.isEmpty())
			{
				Tuple<Message, Connection> res = tryMessagesForConnected(outgoingMessages);
				if(res!=null) // enviou contedo
				{
					Message send = res.getKey(); 
					if (!(R.PC.containsKey(send.getId())))
						R.createNewPC(send.getId(), (send.getData().split(" "))[1], true);
						R.incrSendOutData();
						
					
					this.last_relay_node.put(send.getId(), this.centralitie.get(other));
				}
			}
		}
		
		
	}
	
	
	public void trySendInterestExpiredTTL(List<Message> outgoingMessages)
	{
		List<Tuple<Message, Connection>> outMessages = 
				new ArrayList<Tuple<Message, Connection>>(); 
		
		for (Connection con : getConnections()) {
			
			DTNHost other = con.getOtherNode(getHost());
			sctrRouter othRouter = (sctrRouter)other.getRouter();
			
			
			if (othRouter.isTransferring()) {
				continue; // skip hosts that are transferring
			}
			
			for (Message m : outgoingMessages) outMessages.add(new Tuple<Message, Connection>(m,con));
				
			if(!outgoingMessages.isEmpty())
			{
				Tuple<Message, Connection> res = tryMessagesForConnected(outMessages);
				if(res!=null)
				{
					if(!I.getisSendOutInterest(res.getKey().getId()))
					{
						I.setisSendOutInterest(res.getKey().getId(), true);
						I.setTimeSendOutInterest(res.getKey().getId());
						I.incrSendOutInterests();
					}
					
					Message m = res.getKey();
					this.last_relay_node.put(m.getId(), this.centralitie.get(other));
				}
			}
		}
		
	}
	
	
	public void trysendinterest()
	{
		// ver o novel social do outro n
		
		List<Tuple<Message, Connection>> outgoingMessages = 
			new ArrayList<Tuple<Message, Connection>>(); 
		
		ArrayList<String> idtoRemove = new ArrayList<String>() ;

		Collection<Message> msgCollection = getMessageCollection();
		/* for all connected hosts collect all messages that have a higher
		   probability of delivery by the other host */
		
		List<Cluster> clusters = null;
		
		for (Connection con : getConnections()) {
			DTNHost other = con.getOtherNode(getHost());
			sctrRouter othRouter = (sctrRouter)other.getRouter();
			
			
			if (othRouter.isTransferring()) {
				continue; // skip hosts that are transferring
			}
			
					
			for (Message m : msgCollection) {
				
				
				if (othRouter.hasMessage(m.getId())) {	
					continue; // skip messages that the other one has
				}
				
				
				String split[] = m.getData().split(" ");
				String interestName = split[1];
				
				
				String data = m.getData();
				String[] datas = data.split(" ");
				String name = datas[1];
				
				if (datas[0].equalsIgnoreCase(INTEREST_PACKAGE)) {
					// ver se o n existe na tabela de encontros
					if (this.bloomLC.contains(name)) // Aqui o interesse  satisfeito no n
					{
						if(this.bloomPIT.contains(name)){
							this.bloomPIT.remove(interestName);
							this.PIT.remove(interestName);
							idtoRemove.add(m.getId());						
							I.numberInterestsSastifiedinNode();	
						}
					}
					else if(m.getTo() == other) // chegou ao provedor de contedo
					{
						outgoingMessages.add(new Tuple<Message, Connection>(m,con));
					}
					else if(m.getTo() == null)
					{
						if(this.centralitie.containsKey(other))
						{
							if(this.centralitie.get(other) > this.centralitie.get(this.host))
							{
								clusters = checkcentralitie();
								
								if(!checksameCluster(clusters, this.centralitie.get(this.host), this.centralitie.get(other) ))
								{
									if(this.last_relay_node.containsKey(m.getId()))
									{
										if(this.last_relay_node.get(m.getId()) > this.centralitie.get(other))	
										{
											// envia interesse 
											outgoingMessages.add(new Tuple<Message, Connection>(m,con));
										}
									}
									else {
										outgoingMessages.add(new Tuple<Message, Connection>(m,con));
										}
								}
							}
						}
						
					}
					else if (m.getTo()!= null)
					{
						for (String key : this.socialTieTableOthers.keySet())
						{
							if (this.socialTieTableOthers.get(key).thisidNode == other) // origem 
							{
								if(this.socialTieTableOthers.get(key).idNode == m.getTo()) // para o destino
								{
									if (this.socialTieTable.containsKey(m.getTo()))
											{
												if (this.socialTieTableOthers.get(key).socialvalue > this.socialTieTable.get(m.getTo()).socialvalue)
												{
													outgoingMessages.add(new Tuple<Message, Connection>(m,con));
													// enviar 
												}
											}
								
								}
							}	
						}
						// ver a relao social que o outro n tem com o destino 
						// se tiver uma relao social alta ou seja maior que a dele 
						// se a relao social for maior que a ultima 
							// envia interesse ento 
					}
					

					// se o outro n tiver uma centralidade maior 
					// se tiver mais centralidade aplicar o algoritmo de clustering 
					// se tiver no mesmo cluster no enviar interesse 
					// se tiver noutro cluster ento enviar interesse
						// se last_relay_node == null
						// ento enviar interesse 
						// se last_relay_node != null 
							// ver se a centralidade do n  maior para aquele interesse 
							// se sim enviar interesse
				}
			}
			if(!outgoingMessages.isEmpty())
			{
				Tuple<Message, Connection> res = tryMessagesForConnected(outgoingMessages);
				if(res!=null)
				{
					if(!I.getisSendOutInterest(res.getKey().getId()))
					{
						I.setisSendOutInterest(res.getKey().getId(), true);
						I.setTimeSendOutInterest(res.getKey().getId());
						I.incrSendOutInterests();
					}
					
					Message m = res.getKey();
					this.last_relay_node.put(m.getId(), this.centralitie.get(other));
				}
			}
		}
	}
	
	
	public boolean checksameCluster(List<Cluster> clusters, double thisNode, double otherNode)
	{
		boolean res = false;
		List<DoubleArray> vector; 
		
		for (int i=0; i<clusters.size(); i++)
		{
			vector = clusters.get(i).getVectors();
			if (vector.contains(thisNode) && vector.contains(otherNode))
			{
				return res = true;
			}
			
		}
		
		return res;
	}
	
	
	
	// envia a tabela de conteudos
	
	public void convergeContentTable(DTNHost other)
	{
		MessageRouter otherRouter = other.getRouter();
		
		Map<String, String> otherContents = 
				((sctrRouter)otherRouter).lc;
		
		digest d = new digest();
		
		
		// digest = name
		for(String digest : otherContents.keySet())
		{  
			d.dig.put(digest, SimClock.getTime());
		}
		if(!otherContents.isEmpty()) this.digestTable.put(other, d);		
	}
	
	
	
	
	
	
	// envia a tabela digest
	public void convertDigestTable(DTNHost other)
	{

		MessageRouter otherRouter = other.getRouter();
		
		Map<DTNHost, digest> otherDigest = 
				((sctrRouter)otherRouter).digestTable;
		
		for (DTNHost h :otherDigest.keySet())
		{
			
			if(!(this.host == h))
			{
				for ( String name : otherDigest.get(h).dig.keySet())
				{
					Double timestamp = otherDigest.get(h).dig.get(name);
					
					if(this.digestTable.containsKey(h)) // se existir 
					{
						if(this.digestTable.get(h).dig.containsKey(name))
						{
							if(this.digestTable.get(h).dig.get(name) <  otherDigest.get(h).dig.get(name))
							{
								digest d = new digest();
								d = this.digestTable.get(h);
								d.dig.put(name, otherDigest.get(h).dig.get(name));
								this.digestTable.put(h, d);
							}
						}
						else {
							digest d = new digest();
							this.digestTable.get(h).dig.put(name, otherDigest.get(h).dig.get(name));
						}
					}
					else 
					{
						digest d = new digest();
						d.dig.put(name, otherDigest.get(h).dig.get(name));
						this.digestTable.put(h, d);
					}
				}
			}
			
		}
	}
	
	
	
	@Override
	public sctr replicate() {
		return new sctr(this);
	}
	
	@Override
	protected void dropExpiredMessages() {
		
		Message[] messages = getMessageCollection().toArray(new Message[0]);
		
		List<Tuple<Message, Connection>> outgoingMessages = 
				new ArrayList<Tuple<Message, Connection>>(); 
		
		List<Message> msgs = new ArrayList<Message>();
		for (int i=0; i<messages.length; i++) {
			int ttl = messages[i].getTtl(); 
			if (ttl <= 0) {
				// trysendInterest com flag especial 
				if(messages[i].getData().split(" ")[0].equalsIgnoreCase(INTEREST_PACKAGE)) msgs.add(messages[i]);
				//deleteMessage(messages[i].getId(), true);
			}
		}
		
		if (!outgoingMessages.isEmpty())trySendInterestExpiredTTL(msgs);
	}
	
	@Override
	public void update() {
		super.update();
		if (isTransferring() || !canStartTransfer()) {
			return; // can't start a new transfer
		}
		
	//	this.removeCImessages();
		this.trysendData();	
		this.trysendinterest();
		
	
	}


}


