/* 
 /* 
 * Copyright 2015 Universidade do Minho, GCOM
 * Modified by Paulo Duarte.
 * Released under GPLv3. See LICENSE.txt for details. 
 */

package routing;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import ca.pfv.spmf.algorithms.clustering.kmeans.AlgoKMeans;
import ca.pfv.spmf.patterns.cluster.Cluster;
import report.NDN_Report;
import report.writeReportToFile;
import routing.NDN.DataReport;
import routing.NDN.Interest;
import routing.NDN.InterestReport;
import routing.NDN.digest;
import routing.NDN.socialTie;
import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;
import core.SimClock;

public class sctrRouter extends ActiveRouter {


	/** PeopleShare router's setting namespace ({@value})*/ 
	public static final String NDN_NS = "sctrRouter";
	
	public static final String INTEREST_PACKAGE = "pi";
	public static final String CONTENT_PACKAGE = "pc";
	public static final String CS_TABLE_DIGEST = "ci";
	public static final String FIB_TABLE_DIGEST = "ftd";
	public static final String ORIGINATED_INTEREST_PACKAGE = "pio";
	public static final String INTEREST_NACK_PACKAGE = "inp";
	
	
	/**
	 * Number of seconds in time unit -setting id ({@value}).
	 * How many seconds one time unit is when calculating aging of 
	 * delivery predictions. Should be tweaked for the scenario.*/
	public static final String SECONDS_IN_UNIT_S ="secondsInTimeUnit";
	
	/**
	 * Transitivity scaling constant (beta) -setting id ({@value}).
	 * Default value for setting is {@link #DEFAULT_BETA}.
	 */

	/** the value of nrof seconds in time unit -setting */
	protected int secondsInTimeUnit;
	/** value of beta setting */
	protected double beta;
	
	/** lists of interests*/
	protected Map<String, Interest> li;
	
	// list for incoming node to send data in the interest
	protected Map<String, ArrayList<DTNHost>> lei; 	
	
	/** Forwarding Information Base*/
	protected HashMap<String, Double> FIB;
	protected HashMap<String, PITEntries> PIT;

	/** list of contents*/
	protected Map<String, String> lc;
	
	
	/** PIT with  bloom filter*/
	CBloomFilter<String> bloomPIT;
	/** FIB with  bloom filter*/
	CBloomFilter<String> bloomFIB;
	/** LC with  bloom filter*/
	CBloomFilter<String> bloomLC;
	
	

	boolean LC = false; //LoadContent
	public static final String CONTENT = "content";
	public static final String HOSTS = "nrofHosts";
	public String  contentList = "";
	private int nrofHosts;
	private int id = 0;
	double p_fp = 0.1; // False positive probability
	int n = 100; // Expected number of elements
	
	/** -- Reports*/
	DataReport R;
	InterestReport I;
	writeReportToFile wtf;
	/**Reports -- */
	
	// SCTR
	// <id-node, timestamp>
	protected Map<DTNHost, ArrayList<Double>> encounterVector;
	
	// <Host, <Host, R(), Timestamp>>
	double alpha = 0.5;
	protected Map<DTNHost, socialTie> socialTieTable;
	protected Map<String, socialTie> socialTieTableOthers;
	protected Map<DTNHost, Double> centralitie;
	protected Map<DTNHost, digest> digestTable;
	
	
	public  sctrRouter(Settings s) {
		super(s);		
		Settings sctrRouterSettings= new Settings(NDN_NS);
		contentList = sctrRouterSettings.getSetting(CONTENT);
		nrofHosts = sctrRouterSettings.getInt(HOSTS);
		secondsInTimeUnit = sctrRouterSettings.getInt(SECONDS_IN_UNIT_S);
	
		initRouter();
	}
	
	
	protected sctrRouter(sctrRouter r) {
		super(r);
		this.contentList = r.contentList ;
		this.nrofHosts = r.nrofHosts;
		this.secondsInTimeUnit = r.secondsInTimeUnit;
		this.beta = r.beta;
		
		initRouter();
	}
	
	
	/** Init PIT, FIB, LC*/
	public void initRouter(){
		
		this.li = new HashMap<String, Interest>();
		
		//Create a Counting Bloom filter that has a FP rate of 0.01 when 1000 are inserted
		//and uses 4 Bits for Counting
		
		this.bloomPIT = new CBloomFilter<>(1000, 0.01, 4);
		this.bloomFIB = new CBloomFilter<>(1000, 0.01, 4);
		this.bloomLC = new CBloomFilter<>(1000, 0.01, 4);
		
		this.lei = new HashMap<String, ArrayList<DTNHost>>();
		this.FIB = new HashMap<String, Double>();
		
		//this.FIB = new HashMap<String, ArrayList<DTNHost>>();
		this.lc = new HashMap<String, String>();
		
		this.PIT = new HashMap<String,PITEntries>();
		
		R = DataReport.getInstance();
		I = InterestReport.getInstance();
		
		wtf = writeReportToFile.getInstance();
		
		this.encounterVector = new HashMap<DTNHost, ArrayList<Double>>();
		this.socialTieTable = new HashMap<DTNHost, socialTie>();
		this.socialTieTableOthers = new HashMap<String, socialTie>();
		this.centralitie = new  HashMap<DTNHost, Double>();
		
		this.digestTable = new HashMap<DTNHost, digest>();
		
		if(contentList.equalsIgnoreCase("lista_de_interesses.txt")){
			loadRndContent();
		}
		
	}
		

	@Override
	public sctrRouter replicate() {
		return new sctrRouter(this);
	}
	

	
	private void loadContent(){		
		String line;
		String[] contents, path;
		try{
			File arquivo = new File(contentList);
			FileReader fr = new FileReader(arquivo);
			BufferedReader br = new BufferedReader( fr );
			int n = 0;
			while(br.ready()){		
				line = br.readLine();
				contents = line.split(" "); //contents[0] - host ; contents[1] - content
				path = contents[1].split("/"); 
				if(this.getHost().getAddress() == Integer.valueOf(contents[0])){
					
					if (!this.bloomLC.contains(contents[1])){
						this.lc.put(contents[1], contents[1]);
						this.bloomLC.add(contents[1]);
						//addFIBEntry(contents[1], getHost());
					}
					//addInterest("C" + n, contents[1], 0, false, contents[1]);
					n++;	
					if(NDN_Report.list_of_publishers.containsValue(path[0])){
						continue;
					}
					else{
						NDN_Report.list_of_publishers.put(Integer.valueOf(contents[0]), path[0]);
					}
				}
			}
			br.close();
			fr.close();
		}
		catch(Exception e){
			System.out.println("Erro " + e + " na leitura de " + contentList);
		}
	}

	private void loadRndContent(){		
		String line;
		String[] path;
		boolean temp = false;
		int rndHost = getRandomHost();
		try{
			File arquivo = new File(contentList);
			FileReader fr = new FileReader(arquivo);
			BufferedReader br = new BufferedReader( fr );
			while(br.ready()){		
				line = br.readLine();
				path = line.split("/");
				if(NDN_Report.list_of_publishers.containsValue(path[0])){
					continue;
				}
				else{
					do{
						if(NDN_Report.list_of_publishers.containsKey(rndHost)){
							rndHost = getRandomHost();
					}
						else{
							temp = true;
							NDN_Report.list_of_publishers.put(rndHost, path[0]);
						}
					}while(temp == false);
				}
			}
			br.close();
			fr.close();
		}
		catch(Exception e){
			System.out.println("Erro " + e + " na leitura de " + contentList);
		}
	}
	
	private void uploadContent(){
		String line;
		String[] path;
		try{
			File arquivo = new File(contentList);
			FileReader fr = new FileReader(arquivo);
			BufferedReader br = new BufferedReader( fr );
			int n = 0;
			if(NDN_Report.list_of_publishers.containsKey(this.getHost().getAddress())){
				String domain = NDN_Report.list_of_publishers.get(this.getHost().getAddress());
				while(br.ready()){		
					line = br.readLine();
					path = line.split("/"); 					
					if(domain.equalsIgnoreCase(path[0])){
						//addInterest("I" + n, line, 0, false, line);
						n++;			
					}
				}
				br.close();
				fr.close();
			}

		}
		catch(Exception e){
			System.out.println("Erro " + e + " na leitura de " + contentList);
		}
	}

	public int getRandomHost(){
		Random r = new Random();
		int n = r.nextInt(nrofHosts + 1);
		return n;
	}
	
	 
	
	public void createPITEntrie(String name,ArrayList<DTNHost> h, double t, double timetolive)
	{
		String id = name;
		PITEntries entrie = new PITEntries();
		ArrayList<DTNHost> novaLista = new ArrayList<DTNHost>();
		
		if(!this.PIT.containsKey(name))
		{
			entrie = new PITEntries(id, h, t, timetolive, true); // verificar no outro computador o codigo !!!
			this.PIT.put(name,entrie);
		}

	}
	
	
	
	
	public void addFIBEntry(String contentName, DTNHost host)
	{
		Map<ArrayList<DTNHost>, Double> aux = new HashMap<ArrayList<DTNHost>, Double>() ;
		ArrayList<DTNHost> hostList = new ArrayList<DTNHost>();
		// id interesse(conteudo)
		//DTNHOST (host que sabemos que tem ou possui conhecimento de como chegar ao conteudo
		//Double  a previsibilidade para aquele host)

		if(!FIB.containsKey(contentName)){
			if (host == getHost()){
			hostList.add(host);
			}
			else 
			{
				hostList.add(getHost());
			}
			aux = new HashMap<ArrayList<DTNHost>, Double>();
			aux.put(hostList, 1.0);
			//FIB.put(contentName, aux);
			bloomFIB.add(contentName);
			
		}
	}
	
	
	
	public String getNewId(){
		this.id++;
		String newId = "I" +id;
		if(id == 82) id=0;
		return newId;
	}
	
	

	private String getNameById(String id){
		String path = null;
		for(Interest i: li.values()){
			if(i.getId() == id){
				path = i.getName();
			}
		}
		return path;
	}
	
	/**
	 * Generates a (random) message size
	 * @return message size
	 */
	protected int drawMessageSize() {
	
		Random r = new Random();
		int sizeRangeHigh = 400000; // avg musica
		int sizeRangeLow = 200000; //avg webpages
		//int sizeDiff = NDN_Report.sizeRangeLow == NDN_Report.sizeRangeHigh ? 0 : 
		//	r.nextInt(NDN_Report.sizeRangeHigh - NDN_Report.sizeRangeLow);
		int sizeDiff = sizeRangeLow == sizeRangeHigh ? 0 : 
					r.nextInt(sizeRangeHigh - sizeRangeLow);
		return sizeRangeLow + sizeDiff;
	}
	
	public int getMessageLenght(Message m, String data){
		int lenght = m.getFrom().toString().length() + m.getTo().toString().length() + data.length();
		return lenght;
	}
	
	/**
	 * Called when a connection's state changes. This version doesn't do 
	 * anything but subclasses may want to override this.
	 */
	@Override
	public void changedConnection(Connection con) { }
	
	@Override
	public Message messageTransferred(String id, DTNHost from) {
		Message m = super.messageTransferred(id, from); 
		return m;
	}
	
	// update esquisito 
	// este update chama o update do active router
	// o actuve router chama o update de message router
	// 
	@Override
	public void update() {
		super.update();
		if(!LC){
			LC=true;
			if(contentList.equalsIgnoreCase("lista_de_interesses.txt")){
				uploadContent();
			}
			else
				loadContent();			
		}
		
		//String write = "" + SimClock.getTime() + "	" + I.numberInterestsSastified;
		//wtf.writeNumberInterestsSastified(I);
		//wtf.AvgDelaywithCI(I);
	}
	
	
	public List<Cluster> checkcentralitie()
	{
		
		double myCentralitie;
		double otherCentralitie;
		AlgoKMeans algoKMeans = new AlgoKMeans();  // we request 3 clusters
		int k = 10;
		List<Cluster> clusters = null;
		
	//	myCentralitie = this.centralitie.get(this.host);
		
				// ver se esto no mesmo cluster, se sim ento no enviar 
		
		if(!this.centralitie.isEmpty())
		{
			try {
				clusters = algoKMeans.runAlgorithm(this.centralitie, k);
			} catch (NumberFormatException | IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			

		
		}
		return clusters;
				
	}
	

	public Map<String, Interest> getLI()
	{
		return this.li;
	}
	
	
	/** Insert new interest in BloomPIT, PIT, LI
	 * Define if this router is subscriber or not
	 * @param m, message Interest
	 * @param isSub, true if this router is subscriber, false if isn't subscriber
	 * 
	 * */
	
	public void insertPI(Message m, boolean isSub)
	{
		System.out.println("from: "+ m.getFrom()  + " TO: " + m.getTo());
		ArrayList<DTNHost> newList = new ArrayList<DTNHost>();
	
	
		String data = m.getData(); // obtem os dados
		DTNHost host = m.getFrom(); // obtem a origem do interesse
		
		String idInterest = m.getId();
	
		String[] datas = data.split(" ");
		String name = datas[1];
	
		newList.add(host);
	
		if(this.bloomPIT.contains(name)){ // se existir na PIT
			
			if(!this.PIT.get(name).getAllHosts().contains(host))
			{
				this.PIT.get(name).addHost(host);
			}
						
			if(!I.LI.containsKey(idInterest))
			{
				I.createNewRinterest(idInterest, name);
			}
		
		}
	else { 			
			if(!I.LI.containsKey(idInterest))
			{
				I.createNewRinterest(idInterest, name);
			}
			
			if(this.bloomPIT.add(name)) // cria nova entrada no bloomPIT
			{ 				
				System.out.println("Insert interest  " + (String) name + " em " + m.getFrom()  );
				createPITEntrie(name, newList, SimClock.getTime(), m.getTtl()); // cria nova entrada na PIT
			
			}
	}
	
}
	
	
	

}
